"""
Comparison of splinebox and scipy: Edge Fitting
-----------------------------------------------

This example compares the performance of ``splinebox`` and ``scipy`` when fitting a spline to an image feature, in this case, an integral symbol in a cropped section of an image.
"""

# sphinx_gallery_thumbnail_number = 2

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.optimize
import skimage
import splinebox

# %%
# Plot Styling
# ------------
# To ensure consistent plot aesthetics, let's set some default parameters for Matplotlib. This section can be ignored if you do not need custom plot styling.
mpl.rcParams["image.cmap"] = "gray"
mpl.rcParams["lines.linewidth"] = 2
# mpl.rcParams["axes.prop_cycle"] = cycler.cycler(color=("r",))
scipy_blue = "#0053a6"

# %%
# Load and Display the Image
# --------------------------
# We use a cropped version of the 'text' example image from ``skimage``. This section crops image to the portion containing the integral symbol.
img = skimage.data.text()
img = img[45:78, 300:430]
plt.imshow(img)
plt.show()

# %%
# In order to evaluate how well our spline fits the line, we will have to
# interpolate the pixel values at non-integer position.
# Since the line is black on a white background, our goal is to move the
# spline in a way that minimises the average pixel value under it.
# This is commonly refered to as the image energy.
# To be able to quickly interpolate the pixel values, we fit a bivariate spline to the pixel values.
# The `interpolator` object is callable and can be querried for pixel coordinates.
interpolator = scipy.interpolate.RectBivariateSpline(np.arange(img.shape[0]), np.arange(img.shape[1]), img, s=1)

# %%
# Define Initial Knots
# --------------------
# We'll fit the spline to the integral symbol. Rather than finding initial points automatically, we'll manually select five points roughly spaced along the integral symbol.

initial_knots = np.array([[24, 12], [23, 47], [16, 72], [8, 97], [8, 124]])

# %%
# Create a Spline Using splinebox
# -------------------------------
# We first use ``splinebox`` to create a spline from the selected knots.

M = len(initial_knots)
basis_function = splinebox.B3()
spline = splinebox.Spline(M, basis_function, closed=False)
spline.knots = initial_knots

# %%
# Let's define the parameter values at which we want to sample the spline.
# Here, we chose to sample 50 points inbetween knots.
t = np.linspace(0, M - 1, M * 50)

# %%
# In order to compar the fitted spline to the intial one,
# we save it's positions and knots for plotting later on.
initial_vals = spline(t)
initial_knots = spline.knots


# %%
# Define the Loss Function for splinebox
# --------------------------------------
# Our loss function combines the image energy (to minimize pixel values along the spline) and an internal energy term that ensures smooth, equidistant knots to avoid sharp turns or loops.
# Here, we use the curvilinear reparametrization energy as our internal energy.
# It promotes equidistant spacing of the knots in terms of arc length.
# In practice, this avoids sharp bends and stops the spline from looping/folding back on itself.
# Without it, the image energy would reward the spline for visiting the darkest pixels
# multiple times.
# The parameter ``alpha`` can be used balance the contribution of the image and internal energies.


def loss_function_splinebox(control_points, alpha):
    spline.control_points = control_points.reshape((-1, 2))
    coordinates = spline(t)
    image_energy = np.mean(interpolator(coordinates[:, 0], coordinates[:, 1], grid=False))
    internal_energy = spline.curvilinear_reparametrization_energy()
    return image_energy + alpha * internal_energy


# %%
# Fit the Spline (splinebox)
# --------------------------
# We use ``scipy.optimize.minimize`` to find the best-fitting spline by minimizing the total energy.
# The parameter alpha controls the balance between image energy and internal energy (here emperically set to 500).
initial_control_points = spline.control_points
scipy.optimize.minimize(loss_function_splinebox, initial_control_points.flatten(), args=(500,))

# %%
# Plot the Results (splinebox)
# ----------------------------
# Finally, we plot the initial and fitted splines for comparison.
fitted_vals = spline(t)
fitted_knots = spline.knots

fix, axes = plt.subplots(2, 1, sharex=True, sharey=True)
axes[0].imshow(img)
axes[0].plot(initial_vals[:, 1], initial_vals[:, 0], label="initial spline")
axes[0].scatter(initial_knots[:, 1], initial_knots[:, 0], label="initial knots")
axes[0].legend()
axes[1].imshow(img)
axes[1].plot(fitted_vals[:, 1], fitted_vals[:, 0], label="fitted spline")
axes[1].scatter(fitted_knots[:, 1], fitted_knots[:, 0], label="fitted knots")
axes[1].legend()
plt.suptitle("SplineBox")
plt.tight_layout()
plt.show()

# %%
# Create a Spline Using scipy
# ---------------------------
# Next, we attempt to achieve the same fit using ``scipy``.
# We construct an initial spline with ``scipy.interpolate.make_interp_spline`` using the same knots.
# NOTE: you have to set ``bc_type`` in order to get a spline with the desired
# number of knots.

# Spline order
k = 3
# The parameter value for the knots
t_knots = np.arange(M)
scipy_spline = scipy.interpolate.make_interp_spline(t_knots, initial_knots, k=3, bc_type="natural")

# Save initial values for comparison
initial_vals = scipy_spline(t)
initial_knots = scipy_spline(scipy_spline.t)[k:-k]


# %%
# Define the Loss Function for scipy
# ----------------------------------
# Since scipy does not have a built-in curvilinear reparametrization energy, we calculate it manually.
def loss_function_scipy(control_points, alpha):
    scipy_spline.c = control_points.reshape((-1, 2))
    coordinates = scipy_spline(t)
    image_energy = np.mean(interpolator(coordinates[:, 0], coordinates[:, 1], grid=False))

    # Compute internal energy (curvilinear reparametrization)
    derivative = scipy_spline.derivative()
    integral = scipy.integrate.quad(lambda t: np.linalg.norm(derivative(t)), 0, M - 1)
    length = integral[0]
    c = (length / M) ** 2
    integral = scipy.integrate.quad(lambda t: (np.linalg.norm(derivative(t)) ** 2 - c) ** 2, 0, M - 1)
    internal_energy = integral[0] / length**4

    return image_energy + alpha * internal_energy


# %%
# Fit the Spline (scipy)
# ----------------------
initial_control_points = scipy_spline.c
scipy.optimize.minimize(loss_function_scipy, initial_control_points.flatten(), args=(500,))

# %%
# Plot the Results (scipy)
# ------------------------
# Finally, we plot the initial and fitted splines for the scipy result.
fitted_vals = scipy_spline(t)
fitted_knots = scipy_spline(scipy_spline.t[k:-k])

fix, axes = plt.subplots(2, 1, sharex=True, sharey=True)
axes[0].imshow(img, cmap="gray")
axes[0].plot(initial_vals[:, 1], initial_vals[:, 0], label="initial spline", color=scipy_blue)
axes[0].scatter(initial_knots[:, 1], initial_knots[:, 0], label="initial knots", color=scipy_blue)
axes[0].legend()
axes[1].imshow(img, cmap="gray")
axes[1].plot(fitted_vals[:, 1], fitted_vals[:, 0], label="fitted spline", color=scipy_blue)
axes[1].scatter(fitted_knots[:, 1], fitted_knots[:, 0], label="fitted knots", color=scipy_blue)
axes[1].legend()
plt.suptitle("SciPy")
plt.tight_layout()
plt.show()



"""
Closed interpolating splines
============================

This example demonstrates how the choice of basis function influences the shape of a spline. We use a fixed set of knots and construct splines using different basis functions to observe their effects.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions
import splinebox.spline_curves

# %%
# We define a set of arbitrary 2D points as knots.

knots = np.array([[0, 100], [50, 100], [100, 50], [100, 0], [100, -100], [50, -50], [0, -100], [-100, 0]])

# %%
# We define the parameter values to evaluate the spline for plotting.

t = np.linspace(0, len(knots), 1000)

# %%
# Set up a grid of subplots for visual comparison of different basis functions
# and loop through different basis functions to construct the corresponding splines and plot them.

n, m = 2, 2
fig, axes = plt.subplots(n, m, sharex=True, sharey=True)

for i, (name, basis_function) in enumerate(
    (
        ("B1", splinebox.basis_functions.B1()),
        ("B3", splinebox.basis_functions.B3()),
        ("Exponential", splinebox.basis_functions.Exponential(len(knots))),
        ("CatmullRom", splinebox.basis_functions.CatmullRom()),
    )
):
    M = len(knots)
    curve = splinebox.spline_curves.Spline(M, basis_function, True)
    curve.knots = knots
    discreteContour = curve(t)

    axes[i // n, i % n].scatter(knots[:, 0], knots[:, 1])
    axes[i // n, i % n].plot(discreteContour[:, 0], discreteContour[:, 1])
    axes[i // n, i % n].set_title(name)
    axes[i // n, i % n].set_aspect("equal", adjustable="box")

plt.show()



"""
Measure the curvature of a peptide
==================================

In this example, we measure the curvature of a peptide in an AFM image.

Image credit: Yiwei Zheng, `LBNI <https://www.epfl.ch/labs/lbni/>`_, `EPFL <https://www.epfl.ch/en/>`_
"""

# sphinx_gallery_thumbnail_number = 5

import matplotlib.pyplot as plt
import numpy as np
import scipy
import skan
import skimage
import splinebox
import tifffile

# %%
# 1. Load and Inspect the Data
# ----------------------------
img = tifffile.imread("peptides.tif")

plt.imshow(img, cmap="afmhot")
plt.show()

# %%
# 2. Segmentation and Skeletonization
# -----------------------------------
thresh = skimage.filters.threshold_otsu(img)
mask = img > thresh
skeleton = skimage.morphology.skeletonize(mask)
label_img = skimage.measure.label(skeleton)
label_biggest = np.argmax(np.bincount(label_img.flatten())[1:]) + 1
skeleton = label_img == label_biggest

plt.imshow(skeleton)
plt.show()

# %%
# 3. Select longest path
# ----------------------

graph, coords = skan.csr.skeleton_to_csgraph(skeleton)
coords = np.stack(coords, axis=-1)

dist_matrix, predecessors = scipy.sparse.csgraph.shortest_path(graph, return_predecessors=True)
stop_index, start_index = np.unravel_index(np.argmax(dist_matrix), dist_matrix.shape)
i = start_index
skeleton_points = []
while i != stop_index:
    skeleton_points.append(coords[i])
    i = predecessors[stop_index, i]
skeleton_points = np.array(skeleton_points)

# %%
# 4. Fit Spline
# -------------

M = 20
basis_function = splinebox.basis_functions.B3()
initial_spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=False)
initial_spline.fit(skeleton_points)

t = np.linspace(0, M - 1, M * 100)
initial_vals = initial_spline(t)
initial_knots = initial_spline.knots

plt.imshow(img, cmap="afmhot")
plt.plot(initial_vals[:, 1], initial_vals[:, 0])
plt.scatter(initial_knots[:, 1], initial_knots[:, 0])
plt.show()

# %%
# 5. Refine Spline
# ----------------


def loss_function(control_points, alpha):
    spline.control_points = control_points.reshape((-1, 2))
    coords = spline(t)
    pixel_values = scipy.ndimage.map_coordinates(img, coords.T)
    image_energy = np.mean(pixel_values)
    internal_energy = np.mean(spline(t, derivative=2) ** 2)
    energy = -1 * image_energy + alpha * internal_energy
    return energy


initial_control_points = initial_spline.control_points
spline = initial_spline.copy()
scipy.optimize.minimize(
    loss_function,
    initial_control_points.flatten(),
    args=(0.3,),
    method="Powell",
    tol=0.01,
)

vals = spline(t)
knots = spline.knots

plt.figure()
plt.imshow(img, cmap="afmhot")
plt.plot(initial_vals[:, 1], initial_vals[:, 0], label="initial spline", alpha=0.3)
plt.scatter(initial_knots[:, 1], initial_knots[:, 0], alpha=0.3)
plt.plot(vals[:, 1], vals[:, 0], label="refined spline")
plt.scatter(knots[:, 1], knots[:, 0])
plt.legend()
plt.show()

# %%
# 6. Curvature Comb
# -----------------

total_length = spline.arc_length()
lengths = np.linspace(0, total_length, 200)
t = spline.arc_length_to_parameter(lengths)

vals = spline(t)
curvature = spline.curvature(t)
normals = spline.normal(t)

comb = vals + 500 * curvature[:, np.newaxis] * normals

plt.figure()
plt.imshow(img, cmap="afmhot")
plt.plot(comb[:, 1], comb[:, 0], label="Curvature comb", color="yellowgreen")
for p in range(0, len(t), 3):
    plt.plot([vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]], color="yellowgreen")
plt.plot(vals[:, 1], vals[:, 0], label="spline")
plt.xlim((0, img.shape[1]))
plt.ylim((img.shape[0], 0))
plt.legend()
plt.show()

# %%
# 7. Plot Curvature vs. Length
# ----------------------------

plt.figure()
plt.axhline(0, linestyle="--", color="black", alpha=0.5)
plt.plot(lengths, curvature)
plt.xlabel("length [px]")
plt.ylabel("curvature")
plt.show()



"""
Moving Frames
=============

`Moving frames`_ provide a local coordinate system along a curve. This system can be used to analyze the space around points on the curve, such as extracting orthogonal images from a volume. An example of this application is the Dendrite-Centric Coordinate System.

SplineBox implements two types of moving frames:

1. `Frenet-Serret Frame`_: Defined by the curve's tangent, normal, and binormal vectors. To compute all three vectors, the curve cannot have any straight segments and no inflection points. Additionally, the frame may rotate around the curve.
2. `Bishop Frame`_: A twist-free alternative that eliminates rotations around the curve and is defined on straight segments and at inflections points [Bishop1975]_.

.. _Moving frames: https://en.wikipedia.org/wiki/Moving_frame
.. _Frenet-Serret Frame: https://en.wikipedia.org/wiki/Frenet-Serret_formulas
.. _Bishop Frame: https://www.jstor.org/stable/2319846
"""

import numpy as np
import pyvista
import splinebox

# %%
# 1. Create a Spline
# ^^^^^^^^^^^^^^^^^^

spline = splinebox.Spline(M=4, basis_function=splinebox.B3(), closed=False)
spline.control_points = np.array(
    [
        [0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 1.0, 0.0],
        [1.0, 1.0, 1.0],
        [0.0, 1.0, 1.0],
        [0.0, 0.0, 0.0],
    ]
)

t = np.linspace(0, spline.M - 1, spline.M * 3)

# %%
# 2. Compute the Frenet-Serret Frame
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

frenet_frame = spline.moving_frame(t, method="frenet")

# %%
# 3. Compute the Bishop Frame
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^

bishop_frame = spline.moving_frame(t, method="bishop")

# %%
# 4. Visualize the Frames
# ^^^^^^^^^^^^^^^^^^^^^^^
# We can visualize the Frenet-Serret and Bishop frames using PyVista.
# The following code plots the two frames side-by-side to highlight their differences.
# The Frenet-Serret frame (left) visibly twists along the curve, while the Bishop frame (right) avoids this twist.

# Create a PyVista mesh for the curve
spline_mesh = pyvista.MultipleLines(points=spline(t))

# Add vectors to the mesh for visualization
spline_mesh["frenet0"] = frenet_frame[:, 0] * 0.2
spline_mesh["frenet1"] = frenet_frame[:, 1] * 0.2
spline_mesh["frenet2"] = frenet_frame[:, 2] * 0.2
spline_mesh["bishop0"] = bishop_frame[:, 0] * 0.2
spline_mesh["bishop1"] = bishop_frame[:, 1] * 0.2
spline_mesh["bishop2"] = bishop_frame[:, 2] * 0.2

# Initialize a PyVista plotter
plotter = pyvista.Plotter(shape=(1, 2), border=False)

# Plot the Frenet-Serret frame
plotter.subplot(0, 0)
spline_mesh.set_active_vectors("frenet0")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="blue")
spline_mesh.set_active_vectors("frenet1")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="red")
spline_mesh.set_active_vectors("frenet2")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="green")
plotter.add_mesh(spline_mesh, line_width=10, color="black")

# Plot the Bishop frame
plotter.subplot(0, 1)
spline_mesh.set_active_vectors("bishop0")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="blue")
spline_mesh.set_active_vectors("bishop1")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="red")
spline_mesh.set_active_vectors("bishop2")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="green")
plotter.add_mesh(spline_mesh, line_width=10, color="black", copy_mesh=True)

plotter.link_views()
plotter.show()

# %%
# If we want the Bishop frame to start in a specific orientation, we can specify
# an :code:`initial_vector` in :meth:`splinebox.spline_curves.Spline.moving_frame()`.



"""
Curvature combs
---------------

In this example, we plot the curvature comb of a 2D spline.
Calculating curvature requires computing the second derivative, which can be sensitive to noise when using numerical methods.
Splines are particularly useful in this context because their derivatives can be computed analytically using the chain rule, resulting in smoother and more accurate curvature calculations.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox

# %%
# We begin by generating an arbitrary B3 spline with 7 knots.
# Any twice-differentiable basis function can be used here.

M = 7
spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=False)
spline.knots = np.array([[0, 0], [1, 1], [2.5, 0.5], [3.5, 1.5], [3, 3], [2, 3.5], [1, 2]])

# %%
# Next, we select finely spaced parameter values along the spline for evaluation and plotting.

t = np.linspace(0, M - 1, 1000)

# %%
# To plot the curvature comb, we need to compute the spline's values, curvature, and normal vectors at each t.
# Normal vectors are unit vectors perpendicular to the spline at each point.

vals = spline(t)

curvature = spline.curvature(t)
normals = spline.normal(t)

# %%
# To construct the curvature comb, we scale the normal vectors by the curvature and a factor `d`.
# The factor `d` controls the height of the comb, allowing us to adjust its visual appearance.
# The scaled normal vectors are then added to the spline values.

max_comb_height = 1
d = max_comb_height / np.max(np.abs(curvature))

comb = vals + d * curvature[:, np.newaxis] * normals

# %%
# Finally, we plot the curvature comb. The backbone of the comb and the spline itself are plotted as lines,
# while the "teeth" of the comb are plotted individually using a loop.

plt.plot(comb[:, 1], comb[:, 0], label="Curvature comb", alpha=0.5)
for p in range(0, len(comb), 7):
    plt.plot([vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]], alpha=0.5)
plt.plot(vals[:, 1], vals[:, 0], label="spline")
plt.gca().set_aspect("equal", "box")
plt.legend()
plt.show()



"""
Active Contours
===============

This example demonstrates a basic active contours (snakes) implementation using splinebox. The goal is to segment the astronaut's head in an image by iteratively refining a spline contour.
"""

# sphinx_gallery_thumbnail_number = 4

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy
import skimage
import splinebox.basis_functions
import splinebox.spline_curves

# %%
# Let's load the astronaut example image from skimage.
img = skimage.data.astronaut()

plt.imshow(img)
plt.show()

# %%
# To make the contour stick to the edges of the image, we compute an edge map.
# First, convert the image to grayscale, apply Gaussian smoothing to reduce noise, and then apply the Sobel filter for edge detection.
gray = skimage.color.rgb2gray(img)
smooth = skimage.filters.gaussian(gray, 3, preserve_range=False)
edge = skimage.filters.sobel(smooth)

plt.imshow(edge)
plt.show()


# %%
# To make calculating the edge energy at non-integer locations easier, we use a surface spline on a regular grid.
# This returns a callable object that can be queried for edge energy values.
edge_energy = scipy.interpolate.RectBivariateSpline(
    np.arange(edge.shape[1]), np.arange(edge.shape[0]), edge, kx=2, ky=2, s=1
)


# %%
# We define an internal energy function to regularize the curvature of the spline.
# The internal energy depends on the first and second derivatives of the spline.
def internal_energy(spline, t, alpha, beta):
    return 0.5 * (alpha * spline(t, derivative=1) ** 2 + beta * spline(t, derivative=2) ** 2)


# %%
# Initialize the spline with 30 knots that form a circle around the astronaut's head.
M = 30
s = np.linspace(0, 2 * np.pi, M + 1)[:-1]
y = 100 + 100 * np.sin(s)
x = 220 + 100 * np.cos(s)
knots = np.array([y, x]).T

# %%
# We keep a copy of the initial knots so we can plot them later.
initial_knots = knots.copy()

# %%
# Now, we can construct a B3 spline and set its control points (coefficients)
# using the knots we generated above.
spline = splinebox.spline_curves.Spline(M=M, basis_function=splinebox.basis_functions.B3(), closed=True)
spline.knots = knots

t = np.linspace(0, M, 400)
contour = spline(t)
plt.imshow(img)
plt.scatter(knots[:, 1], knots[:, 0])
plt.plot(contour[:, 1], contour[:, 0])
plt.show()


# %%
# Here, we set the necessary paramters for the active contours algorithm:
#
# * :math:`\alpha` controls the contribution of the first derivative (smoothness) to the internal energy
# * :math:`\beta` controls the contribution of the second derivative (curvature) to the internal energy

alpha = 0
beta = 0.001

# Store intermediate contours
contours = []
# Store the energy values
external_energies = []

# %%
# Nex, we define the energy function to be minimized. It combines the external energy from the edge map
# and the internal energy based on spline smoothness and curvature.


def energy_function(control_points, spline, t, alpha, beta):
    control_points = control_points.reshape((spline.M, -1))
    spline.control_points = control_points
    contour = spline(t)
    contours.append(contour.copy())

    # Compute external energy from the edge map
    edge_energy_value = np.sum(edge_energy(contour[:, 0], contour[:, 1], grid=False))
    external_energies.append(-edge_energy_value)

    # Compute internal energy
    internal_energy_value = np.sum(internal_energy(spline, t, alpha, beta))

    # Total energy to minimize
    return -edge_energy_value + internal_energy_value


# %%
# The active contours approach consists of iteratively updating our control points (coefficients)
# to minimize the energy and find the optimal contour.
initial_control_points = spline.control_points.copy()
result = scipy.optimize.minimize(
    energy_function, initial_control_points.flatten(), method="Powell", args=(spline, t, alpha, beta)
)

# %%
# Inorder to plot the spline as a smooth line, we have to evaluate it
# more densly than just at the knots.
samples = spline(t)

final_knots = spline(np.arange(M))

# %%
# Finaly, we can plot the result.
plt.imshow(img)
plt.scatter(initial_knots[:, 1], initial_knots[:, 0], marker="x", color="black", label="initial knots")
contours = contours[::2000]
colors = matplotlib.colormaps["viridis"](np.linspace(0, 1, len(contours)))
for contour, color in zip(contours, colors):
    plt.plot(contour[:, 1], contour[:, 0], color=color, alpha=0.2)
plt.scatter(final_knots[:, 1], final_knots[:, 0], marker="o", label="final final_knots")
plt.plot(samples[:, 1], samples[:, 0], label="spline")
plt.legend()
plt.show()



"""
Approximating a Noisy Signal
----------------------------

In this example, we add noise to a sinusoidal signal and then approximate it using an exponential spline.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox

# %%
# We generate example data with 100 points by evaluating a sine function
# at equidistant points over the interval from 0 to 15.
# To simulate real-world data, we add Gaussian noise with a standard deviation of 0.3.

N = 100
x = np.linspace(0, 15, N)
values = np.sin(x) + np.random.normal(0, 0.3, N)

# %%
# To approximate this data, we'll use an exponential spline, which can effectively model sinusoidal functions.
# We aim to use the smallest number of knots that still allows the spline to accurately approximate the signal.
# Empirical testing indicates that 8 knots provide a good balance between simplicity and accuracy in this case.

M = 8
basis_function = splinebox.Exponential(M)
spline = splinebox.Spline(M, basis_function, closed=False)

spline.fit(values)

# %%
# To plot the spline we evaluate it at finely spaced parameter values.

ts = np.linspace(0, M - 1, 1000)
spline_y = spline(ts)
spline_x = np.linspace(x.min(), x.max(), len(ts))

plt.scatter(x, values, label="data")
plt.plot(spline_x, spline_y, label="spline")
plt.scatter(np.linspace(x.min(), x.max(), M), spline.knots, label="knots", marker="x", sizes=np.ones(M) * 150)
plt.legend()
plt.show()



import collections
import copy
import json
import math
import warnings

import numba
import numpy as np
import scipy.integrate

import splinebox.basis_functions


def padding_function(knots, pad_length):
    """
    This is the default padding function of splinebox.
    It applies constant padding to the ends of the knots.

    Parameters
    ----------
    knots : numpy array
        The knots to be padded.
    pad_length : int
        The amount of padding at each end.

    Returns
    -------
    padded_knots : numpy array
        Array of padded knots.
    """
    # Add constant padding to the ends
    if knots.ndim == 1:
        knots = knots[:, np.newaxis]
    return np.pad(knots, ((pad_length, pad_length), (0, 0)), mode="edge")


class Spline:
    """
    Base class for the construction of a spline.

    Parameters
    ----------
    M : int
        Number of knots.
    basis_function : :class:`splinebox.basis_functions.BasisFunction`
        The basis function used to construct the spline.
    closed : boolean
        Whether or not the spline is closed, i.e. the two ends are connected.
    control_points : np.array
        The control points of the spline. Optional, can be provided later.
    padding_function : callable
        A function that accepts an array of knots as the first argument and
        the padding size as the second argument. It should return a padded array.
        If `None`, a padded array has to be supplied when setting the `knots`.
        The default is constant padding with the edge values (see :func:`splinebox.spline_curves.padding_function`).
    """

    _wrong_dimension_msg = "It looks like control_points is a 2D array with second dimension different than two. I don't know how to handle this yet."
    _wrong_array_size_msg = (
        "It looks like control_points is neither a 1 nor a 2D array. I don't know how to handle this yet."
    )
    _no_control_points_msg = "This spline doesn't have any coefficients."
    _unimplemented_msg = "This function is not implemented."

    def __init__(self, M, basis_function, closed=False, control_points=None, padding_function=padding_function):
        if basis_function.support <= M:
            self.M = M
        else:
            raise RuntimeError("M must be greater or equal than the spline generator support size.")

        self.basis_function = basis_function
        self._half_support = self.basis_function.support / 2
        # Number of additional knots used for padding the ends
        # of an open spline
        self._pad = math.ceil(self._half_support) - 1
        self.closed = closed
        self.control_points = control_points
        self.padding_function = padding_function

    def _check_control_points(self):
        """
        Most methods require coefficients to be set before they
        can be used. This helper function checks if the coefficients have been
        set.
        """
        if self.control_points is None:
            raise RuntimeError(self._no_control_points_msg)

    def __str__(self):
        closed_str = "closed" if self.closed else "open"
        if self.control_points is None:
            return f"uninitialized {closed_str} {self.basis_function} spline with {self.M} knots"
        else:
            nD = 1 if self.control_points.ndim == 1 else self.control_points.shape[1]
            return f"{closed_str} {nD}D {self.basis_function} spline with {self.M} knots"

    def __repr__(self):
        return f"splinebox.spline_curves.Spline(M={repr(self.M)}, basis_function={repr(self.basis_function)}, closed={repr(self.closed)}, control_points=np.{repr(self.control_points)})"

    def __eq__(self, other):
        return (
            self.M == other.M
            and self.basis_function == other.basis_function
            and self.closed == other.closed
            and np.all(self.control_points == other.control_points)
        )

    @property
    def control_points(self):
        """
        The control points of this spline, i.e. the c[k]
        in equation :ref:`(1) <theory:eq:1>`.
        """
        return self._control_points

    @control_points.setter
    def control_points(self, values):
        if values is not None:
            n = len(values)
            if self.closed and n != self.M:
                raise ValueError(
                    f"The number of control points must match M for a closed spline. You provided {n} control points for a spline with M={self.M}."
                )
            padded_M = self.M + 2 * self.pad
            if not self.closed and n != padded_M:
                raise ValueError(
                    f"Non-closed splines are padded at the ends, i.e. the effective number of control points is M + 2 * (ceil(support/2) - 1). You provided {n} control points for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                )
        self._control_points = values

    @property
    def knots(self):
        """
        The knots of this spline, i.e. the values of the spline
        at :math:`t=0,1,...,M`.
        """
        if self.padding_function is None and not self.closed:
            t = np.arange(-self.pad, self.M + self.pad)
        else:
            t = np.arange(self.M)
        return self(t)

    @knots.setter
    def knots(self, values):
        knots = np.array(values)
        n = len(knots)
        if self.closed:
            if n != self.M:
                raise ValueError(
                    f"You provided {n} knots for a closed spline with M={self.M}. Expected {self.M} knots."
                )
            self.control_points = self.basis_function.filter_periodic(knots)
        else:
            padded_M = self.M + 2 * self.pad
            if self.padding_function is None:
                if n != padded_M:
                    raise ValueError(
                        f"If you do not provide a padding function, i.e. `padding_function=None`, you have to pad the knots at the end to have length M + 2 * (ceil(support/2) - 1) before setting them. You provided {n} knots for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                    )
            else:
                if n != self.M:
                    raise ValueError(f"You provided {n} knots for a spline with M={self.M}. Expected {self.M} knots.")
                knots = self.padding_function(knots, self.pad)
                if len(knots) != padded_M:
                    raise ValueError(
                        f"Incorrect padding. Expected padded to have length {padded_M} instead of {len(knots)}."
                    )

            knots = np.squeeze(knots)

            self.control_points = self.basis_function.filter_symmetric(knots)

    @property
    def basis_function(self):
        """
        The basis function of the spline. Should be an object
        of a specific implementation of the abstract base class
        :class:`splinebox.basis_functions.BasisFunction`.
        """
        return self._basis_function

    @basis_function.setter
    def basis_function(self, value):
        if value.multigenerator:
            raise ValueError(
                "You are trying to construct a Hermite spline using the ordinary `Spline` class. Use the `HermiteSpline` class instead."
            )
        self._basis_function = value

    @property
    def M(self):
        return self._M

    @M.setter
    def M(self, M):
        if hasattr(self, "control_points") and self.control_points is not None and M != self.M:
            # The has attribute is necessary because M is assigned before control_points in the constructor
            raise RuntimeError(
                "M cannot be changed after the control points were set. Create a new spline or set the control_points to None first."
            )
        self._M = M

    @property
    def closed(self):
        return self._closed

    @closed.setter
    def closed(self, closed):
        if hasattr(self, "control_points") and self.control_points is not None and closed != self.closed:
            # The has attribute is necessary because closed is assigned before control_points in the constructor
            raise RuntimeError(
                "closed cannot be changed after the control points were set. Create a new spline or set the control_points to None first."
            )
        self._closed = closed

    @property
    def half_support(self):
        return self._half_support

    @half_support.setter
    def half_support(self, _):
        raise RuntimeError("The half support is determined by the basis function and cannot be set by the user.")

    @property
    def pad(self):
        return self._pad

    @pad.setter
    def pad(self, _):
        raise RuntimeError(
            "The amount of necessary padding is automatically calculated based on the support of the basis function and cannot be changed."
        )

    def copy(self):
        """
        Returns a deep copy of this spline.
        """
        return copy.deepcopy(self)

    def _to_dict(self, version):
        """
        Helper function that creates a dictionary
        representing the spline that can be saved as a json.
        This is implemented separately from :meth:`splinebox.spline_curves.Spline.to_json`
        to allow the :class:`splinebox.spline_curves.HermiteSpline` to inherit this
        conversion only adding the addiontion tangents.

        Paramters
        ---------
        version : int
            The version of the convertion for future compatibility.
        """
        dictionary_representation = {
            "version": version,
            "M": self.M,
            "basis_function": str(self.basis_function),
            "closed": self.closed,
            "control_points": self.control_points.tolist(),
        }
        return dictionary_representation

    def to_json(self, path, version=1):
        """
        Saves the spline as a json file.

        Parameters
        ----------
        path : str or pathlib.Path
            The path where the json file should be saved.
        version : int
            The version of the json file. Default is latest version.
        """
        with open(path, "w") as f:
            json.dump(self._to_dict(version), f, indent=2)

    @classmethod
    def from_json(cls, path):
        """
        Constructs a spline from a json file that was saved using
        :meth:`splinebox.spline_curves.Spline.to_json`.

        Parameters
        ----------
        path : str or pathlib.Path
            Path to the json file.
        """
        with open(path) as f:
            data = json.load(f)
        data = _prepared_dict_for_constructor(data)
        return cls(**data)

    def draw(self, x, y):
        """
        Computes whether a point is inside or outside a closed
        spline on a regular grid of points.

        Parameters
        ----------
        x : numpy array
            A 1D array containing the x values of the grid of points.
        y : numpy array
            A 1D array containing the y values of the grid of points.
        """
        self._check_control_points()

        if self.control_points.ndim != 2 or self.control_points.shape[1] != 2:
            raise RuntimeError("draw() can only be used with 2D curves")

        if not self.closed:
            raise RuntimeError("draw() can only be used with closed curves.")

        xx, yy = np.meshgrid(x, y)
        result = self.is_inside(xx.flatten(), yy.flatten())
        return result.reshape(xx.shape)

    def dtheta(self, t):
        r"""
        Helper function for calculating the winding number.

        `dtheta` is the derivative of the polar coordinate :math:`\theta(t)`

        .. math::
            \theta(t) = arctan \left( \frac{y(t)}{x(t)} \right)

        Differentiation yields:

        .. math::
            \frac{d \theta}{dt} = \frac{1}{r^2} \left( x\frac{dy}{dt} - y\frac{dx}{dt} \right) \text{, where } r^2 = x^2 + y^2
        """
        self._check_control_points()
        t = self._convert_to_array(t)
        r = self(t)
        dr = self(t, derivative=1)
        if r.ndim == 1:
            r = r[np.newaxis, :]
            dr = dr[np.newaxis, :]
        r2 = np.linalg.norm(r, axis=1) ** 2
        if np.any(np.isnan(dr)) or np.isclose(r2, 0):
            # Happens the the spline is not differentiable in this location
            # or when the point is at the origin
            # The solution to return 0 is a bit of a hack but works in practice with
            # the numerical integration
            return np.squeeze(np.zeros(r.shape[0]))
        val = (1.0 / r2) * (r[:, 0] * dr[:, 1] - r[:, 1] * dr[:, 0])
        return np.squeeze(val)

    def is_inside(self, x, y):
        r"""
        Determines if a point with coordinates `x`, `y` is inside the spline.
        Only works for closed 2D curves.

        To determine whether a point is inside or outside the spline, the winding number
        is used:

        .. math::
            wind(\gamma, 0) = \frac{1}{2\pi} \oint_\gamma d\theta = \frac{1}{2\pi} \oint_\gamma \left( \frac{x}{r^2}dy - \frac{y}{r^2}dx \right).

        For a description of :math:`d\theta` check :meth:`splinebox.spline_curves.Spline.dtheta`.

        Parameters
        ----------
        x : numpy.ndarray or float
            x coordinate(s) of point(s)
        y : numpy.ndarray or float
            y coordinate(s) of point(s)

        Returns
        -------
        val : float
            1 if the point is inside, 0.5 if its on the curve and 0 if it is outside the curve.
        """
        if not self.closed:
            raise RuntimeError("isInside() can only be used with closed curves.")
        self._check_control_points()
        if self.control_points.ndim != 2 or self.control_points.shape[1] != 2:
            raise RuntimeError("isInside() can only be used with 2D curves.")

        if isinstance(x, (float, int)):
            x = np.array([x])
        if isinstance(y, (float, int)):
            y = np.array([y])
        if not np.allclose(x.shape, y.shape):
            raise ValueError("x and y need to have the same shape")

        results = np.zeros(x.shape)
        for coord, point in enumerate(np.stack([x, y], axis=-1)):
            spline_copy = self.copy()
            spline_copy.translate(-point)
            winding_number = scipy.integrate.quad(spline_copy.dtheta, 0, spline_copy.M)[0]
            for ref in np.array([[1, 1], [1, 2], [2, 2], [2, 1]]):
                if np.allclose(point, ref):
                    print(ref, winding_number)
            if np.abs(np.abs(winding_number) - 2 * np.pi) < 1e-6:
                results[coord] = 1
            elif np.abs(winding_number) > 1e-6:
                results[coord] = 0.5
            else:
                results[coord] = 0
        results = np.squeeze(results)
        if results.ndim == 0:
            return float(results)
        return results

    def fit(self, points, arc_length_parameterization=False):
        """
        Fit the provided points with the spline using
        least squares.
        For details refer to :ref:`theory/data_approximation:Data approximation`.

        Parameters
        ----------
        points : numpy.ndarray
            The data points that should be fit.
        arc_length_parameterization : bool
            Whether or not to space the knots based on the distance
            between the provided points. This is usefull when the
            points are not equally spaced. Default is `False`.
        """
        if len(points) < self.M + 2 * self.pad:
            raise RuntimeError(
                f"You provided too few points. For the fit to have a unique solution you need to provide at least as many points as your spline has control points (including padding). This spline has {self.M}+{2 * self.pad}(padding) control points but you provided {len(points)} points. Consider providing more points or reducing the number of knots M. If the number of points is equal to M, consider using `spline.knots = points` instead of fitting."
            )

        if arc_length_parameterization:
            raise NotImplementedError
        else:
            t = np.linspace(0, self.M, len(points) + 1)[:-1] if self.closed else np.linspace(0, self.M - 1, len(points))
        tval = self._get_tval(t)
        basis_function_values = self.basis_function(tval, derivative=0)
        self.control_points = np.linalg.lstsq(basis_function_values, points, rcond=None)[0]

    def arc_length(self, stop=None, start=0, epsabs=1e-6, epsrel=1e-6):
        """
        Compute the arc length of the spline between
        the two parameter values specified. If no value for start is give,
        start from the begining of the spline.
        If no value for stop is give, go until the end of the spline.
        When arrays with multiple values are given for start and/or stop,
        an array with all of the arc lengths is returned.

        Parameters
        ----------
        stop : np.array / float (optional)
            Stop point(s) in parameter space.
        start : np.array / float (optional)
            Start point(s) in parameter space.
        epsabs : float (optional)
            Absolute error tolerance. Default is 1e-6.
        epsrel : float (optional)
            Relative error tolerance. Default is 1e-6.
        """
        self._check_control_points()

        if isinstance(stop, collections.abc.Iterable):
            results = np.zeros(len(stop))
            if isinstance(start, collections.abc.Iterable):
                if len(stop) != len(start):
                    raise ValueError(
                        "If you provide array like objects for start and stop, they need to have the same length."
                    )
                for i in range(len(stop)):
                    results[i] = self.arc_length(stop[i], start[i], epsabs=epsabs, epsrel=epsrel)
            else:
                sort_indices = np.argsort(stop)
                sorted_stop = stop[sort_indices]
                for i in range(len(stop)):
                    if i == 0:
                        results[i] = self.arc_length(sorted_stop[i], start, epsabs=epsabs, epsrel=epsrel)
                    else:
                        results[i] = results[i - 1] + self.arc_length(
                            sorted_stop[i], sorted_stop[i - 1], epsabs=epsabs, epsrel=epsrel
                        )
                # Undo the sorting
                results = results[np.argsort(sort_indices)]
            return results

        if stop is None:
            stop = self.M if self.closed else self.M - 1

        if start == stop:
            return 0

        if start > stop:
            start, stop = stop, start

        integral = scipy.integrate.quad(
            lambda t: np.linalg.norm(np.nan_to_num(self(t, derivative=1))),
            start,
            stop,
            epsabs=epsabs,
            epsrel=epsrel,
            maxp1=50,
            limit=100,
        )

        return integral[0]

    def _length_to_parameter_recursion(
        self, s, current_value, lower_bound, upper_bound, intermediate_results=None, atol=1e-4
    ):
        """
        Convert the given arc length s on the curve to a value in parameters space.
        This is done recursively, i.e. check if the point is before or after halfway
        and repeat. It uses binary search.

        Parameters
        ----------
        s : float
            Arc length on the spline.
        current_value : float
            The arc length to the lower bound.
        lower_bound : float
            Lower limit in parameter space.
        upper_bound : float
            Upper limit in parameters space.
        intermediate_results : list
            A list where all computed length parameter pairs are stored.
            This can be used to initialize subsequent conversions more efficiently.
        atol : float
            Absolute precision to which the length is matched.
        """
        self._check_control_points()
        midpoint = lower_bound + (upper_bound - lower_bound) / 2
        midpoint_length = current_value + self.arc_length(lower_bound, midpoint, epsabs=atol)
        if intermediate_results is not None:
            intermediate_results.append((midpoint, midpoint_length))

        if np.isclose(s, midpoint_length, atol=atol, rtol=0):
            return midpoint
        elif s < midpoint_length:
            return self._length_to_parameter_recursion(
                s, current_value, lower_bound, midpoint, intermediate_results, atol
            )
        else:
            return self._length_to_parameter_recursion(
                s, midpoint_length, midpoint, upper_bound, intermediate_results, atol
            )

    def arc_length_to_parameter(self, s, atol=1e-4):
        """
        Convert the arc length `s` to the coresponding value in parameter space.

        Parameters
        ----------
        s : float or np.array
            Length on curve.
        atol : float
            The ablsolute error tolerance.
        """
        self._check_control_points()
        s = self._convert_to_array(s)
        sort_indices = np.argsort(s)
        results = np.zeros_like(s, dtype=float)

        current_value = 0
        lower_bound = 0
        upper_bound = self.M if self.closed else self.M - 1
        intermediate_results = []

        def upper_bound_key_func(x):
            diff = x[1] - s[i]
            if diff >= 0:
                return diff
            else:
                return np.inf

        for i in sort_indices:
            if len(intermediate_results) > 0:
                upper_bound, upper_bound_len = min(intermediate_results, key=upper_bound_key_func)
                if s[i] > upper_bound_len:
                    upper_bound = self.M if self.closed else self.M - 1

            results[i] = self._length_to_parameter_recursion(
                s[i], current_value, lower_bound, upper_bound, intermediate_results, atol=atol
            )

            lower_bound = results[i]
            _, current_value = min(intermediate_results, key=lambda x: abs(x[0] - lower_bound))

        return np.squeeze(results)

    def curvilinear_reparametrization_energy(self, epsabs=1e-6, epsrel=1e-6):
        """
        This energy can be used to enforce equal spacing of the knots.

        Implements equation 25 from [Jacob2004]_.
        In order to make the energy scale invariant,
        we added a factor of (arc length)^-4 to the integral.

        Parameters
        ----------
        epsabs : float
            The absolute accuracy for the integration.
            Default is 1e-6.
            For details see scipy.integrate.quad_.
        epsrel : float
            The relative accuracy for the integration.
            Default is 1e-6.
            For details see scipy.integrate.quad_.

        Returns
        -------
        energy : float
            The curvilinear reparametrization energy of the spline.

        .. _scipy.integrate.quad: https://docs.scipy.org/doc/scipy-1.14.0/reference/generated/scipy.integrate.quad.html
        """
        arc_length = self.arc_length()
        c = (arc_length / self.M) ** 2
        upper_limit = self.M if self.closed else self.M - 1
        integral = scipy.integrate.quad(
            lambda t: (np.linalg.norm(np.nan_to_num(self(t, derivative=1))) ** 2 - c) ** 2,
            0,
            upper_limit,
            epsabs=epsabs,
            epsrel=epsrel,
            maxp1=50,
            limit=100,
        )
        return integral[0] / arc_length**4

    def curvature(self, t):
        """
        Compute the curvature of the spline at position(s) t.
        For splines in 1 and 2 dimensions, the signed curvature
        is returned. Otherwise the unsigned curvature is returned.

        Parameters
        ----------
        t : float or numpy array
            The paramter value(s) at which the curvature
            should be calculated.

        Returns
        -------
        k : float or numpy array
            The curvature value.
        """
        self._check_control_points()
        t = self._convert_to_array(t)
        first_deriv = self(t, derivative=1)
        second_deriv = self(t, derivative=2)
        if first_deriv.ndim == 1:
            # This assumes a uniform spline, i.e. the derivative of t is constant.
            first_deriv = np.stack([first_deriv, np.ones(len(t))], axis=-1)
            second_deriv = np.stack([second_deriv, np.zeros(len(t))], axis=-1)
        norm_first_deriv = np.linalg.norm(first_deriv, axis=1)
        norm_second_deriv = np.linalg.norm(second_deriv, axis=1)
        if first_deriv.shape[1] == 2:
            # We use a different formular for the 2D case to get the signed curvature instead of the
            # the unsigned curvature. This is useful when plotting curvature combs.
            nominator = first_deriv[:, 1] * second_deriv[:, 0] - first_deriv[:, 0] * second_deriv[:, 1]
        else:
            nominator = np.sqrt(
                norm_first_deriv**2 * norm_second_deriv**2 - np.sum(first_deriv * second_deriv, axis=1) ** 2
            )
        k = nominator / norm_first_deriv**3
        return np.squeeze(k)

    def normal(self, t, frame="bishop", initial_vector=None):
        """
        Returns the normal vector for 1D and 2D splines.
        The normal vector points to the right of the spline
        when facing in the direction of increasing t.

        Parameters
        ----------
        t : float or numpy array
            The parameter value(s) for which the normal
            vector(s) are computed.
        frame : "bishop" | "frenet"
            The type of frame used for 3D curves.
        initial_vector : numpy array
            Fixes the initial orientation of the normals at
            position t[0].

        Returns
        -------
        normals : numpy array
            The normal vectors.
        """
        self._check_control_points()
        t = self._convert_to_array(t)
        if self.control_points.ndim != 2:
            raise NotImplementedError(
                "The normal vector is only implemented for curves in 2D and 3D. Your spline's codomain is 1 dimensional."
            )
        if self.control_points.shape[1] == 2:
            first_deriv = self(t, derivative=1)
            normals = (np.array([[0, -1], [1, 0]]) @ first_deriv.T).T
            normals /= np.linalg.norm(normals, axis=1)[:, np.newaxis]
        elif self.control_points.shape[1] == 3:
            frame = self.moving_frame(t, method=frame, initial_vector=initial_vector)
            normals = frame[:, 1:]
        else:
            raise RuntimeError(
                f"The normal vector is only defined for curves in 2D and 3D. Your spline's codomain is {self.control_points.shape[1]} dimensional."
            )
        return normals

    def moving_frame(self, t, method="frenet", initial_vector=None):
        """
        Compute a moving frame (local orthonormal coordinate system) along the spline.

        This method computes either the Frenet-Serret frame or the Bishop frame [#bishop]_ for
        the spline. A moving frame [#movingframe]_ consists of three orthonormal basis vectors at
        each point on the curve. The Frenet-Serret frame is derived from the curve's
        derivatives but may twist around the curve. The Bishop frame eliminates
        this twist, providing a zero-torsion alternative.

        Parameters
        ----------
        t : np.array or float
            A 1D array of parameter values or a single parameter value at which to evaluate the frame.
        method : str, optional
            The type of moving frame to compute. Options are:

            - "frenet": The classical Frenet-Serret frame, based on tangent, normal, and binormal vectors.
            - "bishop": A twist-free frame that requires an initial orientation.

            Default is "frenet".
        initial_vector : np.array or None, optional
            For the Bishop frame, an initial vector that is orthogonal to the tangent
            vector at `t[0]`. This vector determines the initial orientation of the
            basis, which is propagated along the curve without twisting. If None,
            the method computes a suitable initial vector automatically. This
            parameter is ignored when :code:`method="frenet"`.

        Returns
        -------
        frame : np.array
            A 3D numpy array with shape `(len(t), 3, 3)`. The dimensions are:

            - The first axis corresponds to the parameter values in `t`.
            - The second axis contains the three basis vectors at each `t`:
              [tangent, normal, binormal] for "frenet" or equivalent vectors for "bishop".
            - The third axis contains the components of each basis vector in 3D space.

        Raises
        ------
        RuntimeError
            If the spline is not defined in 3D or if the Frenet frame cannot be
            computed due to inflection points, straight segments, or undefined
            tangent/normal vectors.
        ValueError
            If the initial vector for the Bishop frame is not orthogonal to the
            tangent at `t[0]`, or if an invalid `method` is specified.

        Notes
        -----
        - The Frenet frame is not defined at points where the curve has zero curvature,
          such as straight segments or inflection points. In these cases, the Bishop
          frame is recommended.
        - For closed curves, check for discontinuities of the Bishop frame.

        References
        ----------
        .. [#movingframe] `Moving frame <https://en.wikipedia.org/wiki/Moving_frame>`_ on Wikipedia.
        .. [#bishop] Bishop, R. L. (1975). "There is More than One Way to Frame a Curve."
               American Mathematical Monthly, 82(3), 246-251.
        """
        self._check_control_points()

        if self.control_points.ndim != 2 or self.control_points.shape[1] != 3:
            raise RuntimeError("A frame can only be computed for splines in 3D.")

        t = self._convert_to_array(t)

        first_derivative = self(t, derivative=1)
        if first_derivative.ndim == 1:
            first_derivative = first_derivative[np.newaxis]

        frame = np.zeros((len(t), 3, 3))
        frame[:, 0] = first_derivative / np.linalg.norm(first_derivative, axis=-1)[:, np.newaxis]

        if method == "frenet":
            second_derivative = self(t, derivative=2)
            frame[:, 2] = np.cross(first_derivative, second_derivative)
            norm_binormal = np.linalg.norm(frame[:, 2], axis=-1)[:, np.newaxis]
            if np.any(np.isclose(norm_binormal, 0)):
                if np.isclose(norm_binormal[0], 0) or np.isclose(norm_binormal[-1], 0):
                    raise RuntimeError(
                        "The Frenet frame cannot be computed at one or both ends of the spline. This is often due to edge padding of the knots. Try to skip t=0 and t=M-1 or change the padding."
                    )
                raise RuntimeError(
                    "The Frenet frame is not defined for splines with inflection points or straight segments, try the Bishop frame instead."
                )
            frame[:, 2] /= norm_binormal
            frame[:, 1] = np.cross(frame[:, 2], frame[:, 0])
        elif method == "bishop":
            if initial_vector is None:
                tangent = frame[0, 0]
                # Try to do the same as for the Frenet frame
                initial_vector = np.cross(np.cross(tangent, self(t[0], derivative=2)), tangent)
                if np.isclose(np.linalg.norm(initial_vector), 0) or np.any(np.isnan(initial_vector)):
                    initial_vector = np.zeros(3)
                    max_axis = np.argmax(np.abs(tangent))
                    other_axis = (max_axis + 1) % 3
                    initial_vector[max_axis] = tangent[other_axis]
                    initial_vector[other_axis] = -tangent[max_axis]
            initial_vector /= np.linalg.norm(initial_vector)
            if not np.isclose(np.dot(frame[0, 0], initial_vector), 0):
                raise ValueError("The initial vector has to be orthogonal to the tangent at t[0].")
            frame[0, 1] = initial_vector
            frame[0, 2] = np.cross(frame[0, 0], initial_vector)
            for i in range(1, len(t)):
                n = np.cross(frame[i - 1, 0], frame[i, 0])
                norm_n = np.linalg.norm(n)
                if np.isclose(norm_n, 0):
                    frame[i, 1] = frame[i - 1, 1]
                    frame[i, 2] = frame[i - 1, 2]
                else:
                    n /= norm_n
                    phi = np.arccos(np.dot(frame[i - 1, 0], frame[i, 0]))
                    frame[i, 1] = (
                        frame[i - 1, 1] * np.cos(phi)
                        + np.cross(n, frame[i - 1, 1]) * np.sin(phi)
                        + n * np.dot(n, frame[i - 1, 1]) * (1 - np.cos(phi))
                    )
                    frame[i, 2] = (
                        frame[i - 1, 2] * np.cos(phi)
                        + np.cross(n, frame[i - 1, 2]) * np.sin(phi)
                        + n * np.dot(n, frame[i - 1, 2]) * (1 - np.cos(phi))
                    )
        else:
            raise ValueError(f"Unkown method '{method}' for moving frame.")
        return frame

    def __call__(self, t, derivative=0):
        """
        Evalute the spline or one of its derivatives at
        parameter value(s) `t`.

        Parameters
        ----------
        t : numpy.array, float
            A 1D numpy array or a single float value.
        derivative : int
            Can be 0, 1, 2 for the spline, and its
            first and second derivative respectively.
        """
        self._check_control_points()
        t = self._convert_to_array(t)
        # Get values at which the basis functions have to be evaluated
        tval = self._get_tval(t)
        basis_function_values = self.basis_function(tval, derivative=derivative)
        value = np.matmul(basis_function_values, self.control_points)
        return np.squeeze(value)

    def eval(self, t, derivative=0):
        """
        eval is deprecated use :meth:`splinebox.spline_curves.Spline.__call__` instead.
        """
        warnings.warn(
            "`spline.eval(t)` is deprecated and will be removed in v1 use `spline(t)` instead.",
            DeprecationWarning,
            stacklevel=1,
        )
        return self(t, derivative=derivative)

    def _convert_to_array(self, t):
        """
        Helper function that converts the a function input
        to an array. This allows functions to accept int and float
        values in addition to arrays.
        """
        if (isinstance(t, np.ndarray) and t.shape == ()) or (isinstance(t, collections.abc.Iterable) and len(t) == 1):
            warnings.warn(
                "Starting from version 1.0 splinebox will return an array of length 1 if the input is length 1 instead of returning a single value. For details see https://github.com/EPFL-Center-for-Imaging/splinebox/issues/53.",
                FutureWarning,
                stacklevel=2,
            )
        if not isinstance(t, collections.abc.Iterable):
            t = np.array([t])
        elif isinstance(t, np.ndarray) and t.shape == ():
            # Array with only one element, e.g. np.array(0.)
            t = np.array([t.item()])
        elif not isinstance(t, np.ndarray):
            t = np.array(t)
        return t

    def _get_tval(self, t):
        """
        This is a helper method for `__call__`. It is its own method
        to allow :class:`splinebox.spline_curves. HermiteSpline` to
        overwrite the `__call__` method using `_get_tval`.
        It is also used in :meth:`splinebox.spline_curves.Spline.fit`
        """
        if self.closed:
            # all knot indices
            k = np.arange(self.M)
            # output array for the helper function _wrap_index
            tval = np.full((len(t), len(k)), np.nan)
            # compute the positions at which the basis functions have to be evaluated
            # and save them in tval
            self._wrap_index(t, k, self.half_support, self.M, tval)
        else:
            # take into account the padding with additional basis functions
            # for non-closed splines
            k = np.arange(self.M + 2 * self.pad) - self.pad
            k = k[np.newaxis, :]
            t = t[:, np.newaxis]
            # positions at which the basis functions have to be evaluated
            tval = t - k
        return tval

    @staticmethod
    @numba.guvectorize(
        [(numba.float64[:], numba.float64[:], numba.float64, numba.int64, numba.float64[:, :])], "(n),(m),(),()->(n,m)"
    )
    def _wrap_index(ts, ks, half_support, M, wrapped_tval):  # pragma: no cover
        """
        Fill the wrapped_tval array whenever a value t
        is affected by the basis function at knot k, taking
        into account that basis functions at the begging/end
        affect positions on the opposite end for closed splines.
        """
        outside_tvalue = half_support + 1
        for i, k in enumerate(ks):
            for j, t in enumerate(ts):
                if t >= M + k - half_support:
                    # t is close enough to the end to be affected by
                    # the k-th basis function from the beginning
                    wrapped_tval[j, i] = t - M - k
                elif t <= half_support - (M - k):
                    # t is close enough to the beginning to be affected
                    # by the k-th basis function, which is close to the end
                    wrapped_tval[j, i] = t + M - k
                elif t > k + half_support or t < k - half_support:
                    # t is outside the support of the k-th basis function
                    # this can be any value outside the support
                    wrapped_tval[j, i] = outside_tvalue
                else:
                    wrapped_tval[j, i] = t - k

    def _control_points_centroid(self):
        """
        Helper method for :meth:`splinebox.spline_curves.Spline.scale`
        and :meth:`splinebox.spline_curves.Spline.rotate`.
        Computes the centroid of the coefficients.
        """
        self._check_control_points()
        return np.mean(self.control_points, axis=0)

    def translate(self, vector):
        """
        Translates the spline by a `vector`.

        Parameters
        ----------
        vector : numpy.ndarray
            Displacement vector added to the coefficients.
        """
        self._check_control_points()
        self.control_points = self.control_points + vector

    def scale(self, scaling_factor):
        """
        Enlarge or shrink the spline by the provided factor.
        """
        self._check_control_points()
        centroid = self._control_points_centroid()
        self.translate(-centroid)
        self.control_points *= scaling_factor
        self.translate(centroid)

    def rotate(self, rotation_matrix, centred=True):
        """
        Rotate the spline with the provided rotation matrix.

        Parameters
        ----------
        rotation_matrix : numpy.ndarray
            The rotation matrix applied to the spline.
        """
        self._check_control_points()
        if self.control_points.ndim == 1:
            raise RuntimeError("1D splines can not be rotated.")

        if centred:
            centroid = self._control_points_centroid()
            self.translate(-centroid)

        for k in range(len(self.control_points)):
            self.control_points[k] = np.matmul(rotation_matrix, self.control_points[k])

        if centred:
            self.translate(centroid)

    def distance(self, point, return_t=False):
        """
        Computes the distance of point from the spline.

        Parameters
        ----------
        point : numpy.array
            Array with the coordinates of the point.
        return_t : bool
            Whether to return the paramter t of the spline.
            `spline(t)` gives the location on the spline
            closest to the point.

        Returns
        -------
        distance : float
            The distance between the point and the spline.
        t : float
            Only returned if `return_t=True`. This is the parameter
            corresponding to the location on the spline closest
            to the point.
        """
        self._check_control_points()
        if self.control_points.ndim == 1:
            raise RuntimeError("Cannot compute distance for 1D splines.")

        max_t = self.M if self.closed else self.M - 1
        t = np.linspace(0, max_t, self.M * 10)
        points_on_spline = self(t)
        distances = np.linalg.norm(points_on_spline - point[np.newaxis], axis=-1)
        t_initial = t[np.argmin(distances)]

        def _distance(t):
            return np.linalg.norm(self(t) - point)

        result = scipy.optimize.minimize(_distance, np.array((t_initial,)), bounds=((0, max_t),))

        min_distance = np.linalg.norm(self(result.x) - point)

        if return_t:
            return (min_distance, result.x)

        return min_distance

    def mesh(
        self,
        radius=None,
        step_t=0.1,
        step_angle=5,
        mesh_type="surface",
        cap_ends=False,
        frame="bishop",
        initial_vector=None,
    ):
        """
        Create a 3D mesh around the spline curve.

        This method generates a mesh surrounding the spline, with the distance from
        the spline controlled by the `radius` parameter. The mesh can be either a
        surface mesh with triangular cells or a volume mesh with tetrahedral cells.
        The orientation of the mesh is determined using either the Frenet-Serret or
        Bishop frame.

        Parameters
        ----------
        radius : float or callable, optional
            Defines the distance from the spline to the mesh surface. Can be:

            - A float for a constant radius.
            - A callable function that takes the spline parameter `t` and the polar
              angle in the normal plane as arguments and returns a float.

            If None, the mesh will follow the spline without an offset. Default is None.
        step_t : float, optional
            Step size for the spline parameter `t`, controlling the resolution along
            the curve. Smaller values increase mesh resolution. Default is 0.1.
        step_angle : float, optional
            Step size for the polar angle (in degrees), controlling the resolution
            around the spline. Smaller values increase mesh resolution. Default is 5.
        mesh_type : str, optional
            Specifies the type of mesh to create:

            - "surface": A surface mesh with triangular cells.
            - "volume": A volume mesh with tetrahedral cells.

            Default is "surface".
        cap_ends : bool, optional
            If True, the ends of an open surface mesh are capped with orthogonal
            planes. Ignored for closed splines or volume meshes. Default is False.
        frame : str, optional
            The frame to use for orientation of the mesh:
            - "frenet": Uses the Frenet-Serret frame.
            - "bishop": Uses the Bishop frame, requiring an `initial_vector`.
            See :meth:`splinebox.spline_curves.moving_frame`. Default is "bishop".
        initial_vector : numpy array or None, optional
            For the Bishop frame, an initial vector that defines the orientation of
            the frame at the start of the spline (`t[0]`). This vector must be
            orthogonal to the tangent at `t[0]`. Ignored for the Frenet frame. If
            None, a suitable initial vector is computed automatically. Default is None.

        Returns
        -------
        points : numpy array
            A 2D array of shape `(N, 3)`, where `N` is the number of mesh points.
            Each row represents the (x, y, z) coordinates of a point in the mesh.
        connectivity : numpy array
            A 2D array of shape `(M, K)`, where `M` is the number of elements in the
            mesh and `K` is the number of vertices per element (3 for surface meshes
            and 4 for volume meshes). Each row contains the indices of `points` that
            form an element.

        Raises
        ------
        NotImplementedError
            If the spline is not defined in 3D, as meshes are only supported for 3D splines.

        Notes
        -----
        - Surface meshes are useful for visualization, while volume meshes are
          typically used in simulations and finite element analysis.
        - For open splines, capping the ends (`cap_ends=True`) creates closed surfaces,
          which may be useful for some applications.
        - The Bishop frame is recommended for curves with inflection points or
          straight segments where the Frenet frame is undefined.
        - When radius is callable, the Bishop frame is recommend to avoid "drift" of the
          polar angle around the curve.

        Examples
        --------
        Create a surface mesh with constant radius:

        >>> points, connectivity = spline.mesh(radius=0.5, step_t=0.1, step_angle=10, mesh_type="surface")

        Create a volume mesh with variable radius:

        >>> def radius_function(t, angle):
        >>>     return 0.5 + 0.2 * np.sin(np.radians(angle))
        >>> points, connectivity = spline.mesh(radius=radius_function, mesh_type="volume")
        """
        if self.control_points.ndim != 2 or self.control_points.shape[1] != 3:
            raise NotImplementedError("Meshes are only implemented for splines in 3D.")
        t = np.arange(0, self.M if self.closed else self.M - 1 + step_t, step_t)
        if radius is None or radius == 0:
            points = self(t)
            connectivity = np.stack((np.arange(len(points)), np.arange(len(points)) + 1), axis=-1)
            if self.closed:
                # Connect end to the beginning
                connectivity[-1, -1] = 0
            else:
                connectivity = connectivity[:-1]
        else:
            _radius = (lambda t, phi: np.full(t.shape, radius)) if not callable(radius) else radius

            phi = np.arange(0, 360, step_angle)

            if mesh_type == "surface":
                phiphi, tt = np.meshgrid(phi, t)
                tt = tt.flatten()
                phiphi = phiphi.flatten()
                centers = self(tt.flatten())
                rr = _radius(tt, phiphi)
                normals = self.normal(t, frame=frame, initial_vector=initial_vector)

                n_angles = len(phi)
                n_t = len(t)
                normals = (
                    np.repeat(normals[:, 0], n_angles, axis=0) * np.sin(np.deg2rad(phiphi))[:, np.newaxis]
                    + np.repeat(normals[:, 1], n_angles, axis=0) * np.cos(np.deg2rad(phiphi))[:, np.newaxis]
                )

                points = centers + rr[:, np.newaxis] * normals
                n_points = len(points)
                connectivity = self._surface_mesh_connectivity(self.closed, n_angles, n_t, n_points)
                if cap_ends and not self.closed:
                    points = np.concatenate((centers[0].reshape((1, -1)), points, centers[-1].reshape((1, -1))), axis=0)
                    start_connectivity = np.zeros((n_angles, 3), dtype=int)
                    start_connectivity[:, 1] = np.arange(1, n_angles + 1)
                    start_connectivity[:, 2] = np.roll(start_connectivity[:, 1], -1)
                    end_connectivity = np.zeros((n_angles, 3), dtype=int)
                    end_connectivity[:, 0] = n_points + 1
                    end_connectivity[:, 1] = np.arange(n_points, n_points - n_angles, -1)
                    end_connectivity[:, 2] = np.roll(end_connectivity[:, 1], -1)
                    connectivity = np.concatenate((start_connectivity, connectivity + 1, end_connectivity))

            elif mesh_type == "volume":
                phiphi, tt = np.meshgrid(phi, t)
                rr = _radius(tt, phiphi)
                # Add columns for the center points
                rr = np.hstack((np.zeros((rr.shape[0], 1)), rr))
                tt = np.hstack((tt[:, 0][:, np.newaxis], tt))
                phiphi = np.hstack((phiphi[:, 0][:, np.newaxis], phiphi))

                tt = tt.flatten()
                phiphi = phiphi.flatten()
                rr = rr.flatten()

                centers = self(tt.flatten())
                normals = self.normal(t, frame=frame, initial_vector=initial_vector)

                n_angles = len(phi)
                n_t = len(t)
                normals = (
                    np.repeat(normals[:, 0], n_angles + 1, axis=0) * np.sin(np.deg2rad(phiphi))[:, np.newaxis]
                    + np.repeat(normals[:, 1], n_angles + 1, axis=0) * np.cos(np.deg2rad(phiphi))[:, np.newaxis]
                )
                points = centers + rr[:, np.newaxis] * normals
                n_points = len(points)

                connectivity = self._volume_mesh_connectivity(self.closed, n_angles, n_t, n_points)

        return points, connectivity

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _surface_mesh_connectivity(closed, n_angles, n_t, n_points):  # pragma: no cover
        if closed:
            connectivity = np.zeros((2 * n_angles * n_t, 3), dtype=numba.int64)
        else:
            connectivity = np.zeros((2 * n_angles * (n_t - 1), 3), dtype=numba.int64)
        face = 0
        for i in range(n_t if closed else n_t - 1):
            for j in range(n_angles):
                connectivity[face] = [
                    i * n_angles + j,
                    ((i + 1) * n_angles + j) % n_points,
                    ((i + 1) * n_angles + (j + 1) % n_angles) % n_points,
                ]
                face += 1
                connectivity[face] = [
                    i * n_angles + j,
                    ((i + 1) * n_angles + (j + 1) % n_angles) % n_points,
                    (i * n_angles + (j + 1) % n_angles) % n_points,
                ]
                face += 1
        return connectivity

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _volume_mesh_connectivity(closed, n_angles, n_t, n_points):  # pragma: no cover
        if closed:
            connectivity = np.zeros((3 * n_angles * n_t, 4), dtype=numba.int64)
        else:
            connectivity = np.zeros((3 * n_angles * (n_t - 1), 4), dtype=numba.int64)
        vol = 0
        for i in range(n_t if closed else n_t - 1):
            for j in range(1, n_angles + 1):
                connectivity[vol] = [
                    i * (n_angles + 1) + j,
                    ((i + 1) * (n_angles + 1) + j) % n_points,
                    ((i + 1) * (n_angles + 1) + 1 + j % n_angles) % n_points,
                    (i + 1) * (n_angles + 1) % n_points,
                ]
                vol += 1
                connectivity[vol] = [
                    i * (n_angles + 1) + j,
                    ((i + 1) * (n_angles + 1) + 1 + j % n_angles) % n_points,
                    (i * (n_angles + 1) + 1 + j % n_angles) % n_points,
                    i * (n_angles + 1),
                ]
                vol += 1
                connectivity[vol] = [
                    i * (n_angles + 1) + j,
                    ((i + 1) * (n_angles + 1) + 1 + j % n_angles) % n_points,
                    (i + 1) * (n_angles + 1) % n_points,
                    i * (n_angles + 1),
                ]
                vol += 1
        return connectivity


class HermiteSpline(Spline):
    """
    Class for the construction of a Hermite spline.
    It inherits from :class:`splinebox.spline_curves.Spline`.
    Here, we only document the additional methods and attributes.
    For information on the inherited methods and attributes refere to the
    documentation of :class:`splinebox.spline_curves.Spline`.

    Parameters
    ----------
    M : int
        Number of control points.
    basis_function : :class:`splinebox.basis_functions.BasisFunction`
        The basis function used to construct the spline. The :class:`multigenerator <splinebox.basis_functions.BasisFunction>`
        attribute has to be true and the :func:`__call__ <splinebox.basis_functions.BasisFunction.__call__>` method has to return two values
        instead of one.
    closed : boolean
        Whether or not the spline is closed, i.e. the two ends are connected.
    """

    _coef_tangent_mismatch_msg = "It looks like control_points and tangents have different shapes."
    _no_tangents_msg = "This spline doesn't have any tangents."

    def __init__(
        self, M, basis_function, closed=False, control_points=None, tangents=None, padding_function=padding_function
    ):
        super().__init__(M, basis_function, closed, control_points=control_points, padding_function=padding_function)
        self.tangents = tangents

    def _check_control_points_and_tangents(self):
        """
        Most methods require coefficients to be set before they
        can be used. This helper function checks if the coefficients have been
        set.
        """
        self._check_control_points()
        if self.tangents is None:
            raise RuntimeError(self._no_tangents_msg)

    def __repr__(self):
        return f"splinebox.spline_curves.HermiteSpline(M={repr(self.M)}, basis_function={repr(self.basis_function)}, closed={repr(self.closed)}, control_points=np.{repr(self.control_points)}, tangents=np.{repr(self.tangents)})"

    def __eq__(self, other):
        return super().__eq__(other) and np.all(self.tangents == other.tangents)

    @property
    def tangents(self):
        """
        The tangents of this spline.
        """
        return self._tangents

    @tangents.setter
    def tangents(self, values):
        if values is not None:
            n = len(values)
            if self.closed and n != self.M:
                raise ValueError(
                    f"The number of tangents must match M for a closed spline. You provided {n} tangents for a spline with M={self.M}."
                )
            padded_M = self.M + 2 * self.pad
            if not self.closed and n != padded_M:
                raise ValueError(
                    f"Non-closed splines are padded at the ends, i.e. the effective number of tangents is M + 2 * (ceil(support/2) - 1). You provided {n} tangents for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                )
        self._tangents = values

    @property
    def basis_function(self):
        return self._basis_function

    @basis_function.setter
    def basis_function(self, value):
        if not value.multigenerator:
            raise ValueError(
                "It looks like you are trying to use a single generator to build a multigenerator spline model."
            )
        self._basis_function = value

    def fit(self, points, arc_length_parameterization=False):
        if len(points) < 2 * (self.M + 2 * self.pad):
            raise RuntimeError(
                f"You provided too few points. For a unique solution you need to provide at least 2*({self.M}+{2 * self.pad}) points to match the number of control points and tangents (including padding). You provided {len(points)} points. Consider providing more points or reducing the number of knots M."
            )
        if len(points) < self.M:
            raise RuntimeError(
                "You provided fewer data points than your spline has knots. For the fit to have a unique solution you need to provide at least as many data points as your spline has knots. Consider adding more data or reducing the number of knots M."
            )
        if arc_length_parameterization:
            raise NotImplementedError
        else:
            t = np.linspace(0, self.M, len(points) + 1)[:-1] if self.closed else np.linspace(0, self.M - 1, len(points))
        tval = self._get_tval(t)
        basis_function_values = self.basis_function(tval, derivative=0)
        basis_function_values = np.concatenate([basis_function_values[0], basis_function_values[1]], axis=1)
        solution = np.linalg.lstsq(basis_function_values, points, rcond=None)[0]
        half = self.M if self.closed else self.M + 2 * self.pad
        self.control_points = solution[:half]
        self.tangents = solution[half:]

    def __call__(self, t, derivative=0):
        self._check_control_points_and_tangents()
        t = self._convert_to_array(t)

        tval = self._get_tval(t)
        basis_function_values = self.basis_function(tval, derivative=derivative)
        value = np.matmul(basis_function_values[0], self.control_points) + np.matmul(
            basis_function_values[1], self.tangents
        )
        return np.squeeze(value)

    def scale(self, scaling_factor):
        self._check_control_points_and_tangents()
        Spline.scale(self, scaling_factor)
        self.tangents *= scaling_factor

    def rotate(self, rotation_matrix, centred=True):
        self._check_control_points_and_tangents()

        Spline.rotate(self, rotation_matrix, centred=centred)

        for k in range(len(self.tangents)):
            self.tangents[k] = np.matmul(rotation_matrix, self.tangents[k])

    def _to_dict(self, version):
        dictionary_representation = super()._to_dict(version)
        dictionary_representation["tangents"] = self.tangents.tolist()
        return dictionary_representation


def _prepared_dict_for_constructor(data):
    """
    Helper function that processes the dictionaries loaded from
    json files. It ensure all of the values are valid and prepares
    a dictionary that can be passed to the constructor using `**`.

    Parameters
    ----------
    data : dict
        The dictionary from the json file.
    """
    if not isinstance(data["version"], int):
        raise ValueError("version has to be an integer.")

    if not isinstance(data["M"], int):
        raise ValueError("M has to be an integer.")

    data["basis_function"] = splinebox.basis_functions.basis_function_from_name(data["basis_function"], M=data["M"])

    data["closed"] = str(data["closed"]).lower()
    true_strings = ["true", "1", "t", "y", "yes"]
    false_strings = ["false", "0", "f", "n", "no"]
    if data["closed"] not in true_strings and data["closed"] not in false_strings:
        raise ValueError(f"closed should be a string that can be interpreted as a boolean not {data['closed']}.")
    data["closed"] = data["closed"] in true_strings

    if "control_points" in data:
        data["control_points"] = np.array(data["control_points"])
    if "tangents" in data:
        data["tangents"] = np.array(data["tangents"])

    # The version of the json file is only required for parsing
    del data["version"]

    return data


def splines_to_json(path, splines, version=1):
    """
    Saves multiple splines in a single json file.

    Parameters
    ----------
    path : str or pathlib.Path
        The path where the json should be saved.
    splines : iterable
        For instance a list of :class:`splinebox.spline_curves.Spline`
        and :class:`splinebox.spline_curves.HermiteSpline` objects.
    version : int
        The version used to produce the json file.
    """
    dicts = []

    for spline in splines:
        dicts.append(spline._to_dict(version))

    with open(path, "w") as f:
        json.dump(dicts, f, indent=2)


def splines_from_json(path):
    """
    Loades multiple splines from a json file generated using
    :func:`splinebox.spline_curves.splines_to_json`.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the json file.

    Returns
    -------
    splines : list
        A list of :class:`splinebox.spline_curves.Spline` and
        :class:`splinebox.spline_curves.HermiteSpline` objects.
    """
    splines = []
    with open(path) as f:
        data = json.load(f)

    if isinstance(data, dict):
        # this is a json file of a single spline
        data = [data]

    for spline_data in data:
        spline_data = _prepared_dict_for_constructor(spline_data)

        if spline_data["basis_function"].multigenerator:
            # This is a basis function for a Hermite spline
            splines.append(HermiteSpline(**spline_data))
        else:
            splines.append(Spline(**spline_data))

    return splines



import matplotlib.pyplot as plt
import numpy as np
import splinebox

spline = splinebox.Spline(M=10, basis_function=splinebox.B3(), closed=False)

delta = np.pi / spline.M
theta = np.linspace(-delta, np.pi + delta, spline.M + 2)
control_points = np.stack([np.cos(theta), np.sin(theta)], axis=-1)

spline.control_points = control_points

t = np.linspace(0, spline.M - 1, 1000)
vals = spline(t)

plt.figure(figsize=(6, 3))
plt.scatter(spline.control_points[1:-1, 0], spline.control_points[1:-1, 1], label="control points")
plt.plot(vals[:, 0], vals[:, 1], label="spline")
plt.gca().set_aspect("equal", "box")
plt.legend()
plt.tight_layout()
plt.show()



"""
Saving and loading splines
--------------------------
"""

import sys

import numpy as np
import splinebox.basis_functions
import splinebox.spline_curves

# %%
# We start by creating a random spline.

spline = splinebox.spline_curves.Spline(
    M=5, basis_function=splinebox.basis_functions.B3(), closed=True, control_points=np.random.rand(5, 3)
)

# %%
# Let's save the spline:

spline.to_json("spline.json")

# %%
# Here is what the json file looks like:

with open("spline.json") as f:
    sys.stdout.write(f.read())

# %%
# Next, we will create a new spline based on the json file.

loaded_spline = splinebox.spline_curves.Spline.from_json("spline.json")

# %%
# You can also save multiple splines in a single json file.

splines = []
for _ in range(3):
    spline = splinebox.spline_curves.Spline(
        M=4, basis_function=splinebox.basis_functions.B3(), closed=True, control_points=np.random.rand(4, 1)
    )
    splines.append(spline)

splinebox.spline_curves.splines_to_json("splines.json", splines)

# %%
# Here is what a json file with multiple splines looks like:

with open("splines.json") as f:
    sys.stdout.write(f.read())

# %%
# Lastly, we load multiple splines from a single json file.

splines = splinebox.spline_curves.splines_from_json("splines.json")
print(splines)



"""
Tracking the Undulating Motion of C. elegans
============================================
This example demonstrates how to track the characteristic undulating pattern of a moving C. elegans using splinebox.

Data Source: `WormSwin <https://zenodo.org/records/7456803>`_

Data Used: csb-1_dataset 24_2_1_1

Preprocessing:

* Background: Maximum projections of frames 75-187 (to account for minor movements).
* Background Subtraction: Subtracted from each frame.
* Cropping: Frames 141-187, focusing on a single worm centered at the bottom.
* Image Adjustment: Rotation, contrast adjustment, and conversion to 8-bit.
"""

# sphinx_gallery_thumbnail_number = 3

import matplotlib.animation
import matplotlib.pyplot as plt
import numpy as np
import skimage
import splinebox

# %%
# Load the data and inspect it
# ----------------------------

stack = skimage.io.imread("celegans.tif")


def _update0(i):
    mpl_img.set_array(stack[i])
    return (mpl_img,)


fig, ax = plt.subplots(figsize=(7, 3))
mpl_img = ax.imshow(stack[0], cmap="Greys_r")
ax.set(xlim=(0, stack.shape[2]), ylim=(0, stack.shape[1]))
animation = matplotlib.animation.FuncAnimation(fig, _update0, len(stack), interval=100, blit=True)
plt.show()

# %%
# Segmentation of the Worm
# ------------------------
# Next, we segment the worm using Otsu's thresholding method to prepare for spline fitting.

thresh = skimage.filters.threshold_otsu(stack)
mask = stack < thresh


def _update1(i):
    mpl_img.set_array(stack[i])
    mpl_mask.set_array(mask[i])
    return (mpl_img, mpl_mask)


fig, ax = plt.subplots(figsize=(7, 3))
mpl_img = ax.imshow(stack[0], cmap="Greys_r")
mpl_mask = ax.imshow(mask[0], cmap="Reds", alpha=0.5)
ax.set(xlim=(0, stack.shape[2]), ylim=(0, stack.shape[1]))
animation = matplotlib.animation.FuncAnimation(fig, _update1, len(stack), interval=100, blit=True)
plt.show()

# %%
# Fitting Splines to the Worm's Shape
# -----------------------------------
# We fit a spline to the worm in each frame. The spline is created by skeletonizing the worm and then ordering the skeleton points based on their x-position. For simplicity, this method assumes that the worm is moving from left to right. In more complex scenarios, a graph-based approach (e.g., using the skan package) would be more appropriate for ordering the points.
# We use seven knots in our spline, balancing accuracy and robustness against segmentation imperfections. An exponential basis function is selected for the spline to enable the generation of trigonometric functions.

M = 7
basis_function = splinebox.basis_functions.Exponential(M)

splines = []
for i in range(len(mask)):
    label_img = skimage.measure.label(mask[i])
    label_biggest = np.argmax(np.bincount(label_img.flatten())[1:]) + 1
    mask[i] = label_img == label_biggest
    mask[i] = skimage.morphology.binary_opening(mask[i])
    mask[i] = skimage.morphology.remove_small_holes(mask[i], area_threshold=7)
    skeleton = skimage.morphology.skeletonize(mask[i])
    spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=False)
    coords = np.stack(np.where(skeleton), axis=-1)
    coords = coords[np.argsort(coords[:, 1])]
    spline.fit(coords)
    splines.append(spline)

t = np.linspace(0, M - 1, 500)


def _update2(i):
    mpl_img.set_array(stack[i])
    vals = splines[i](t)
    mpl_line.set_data(vals[:, 1], vals[:, 0])
    return (mpl_img, mpl_line)


fig, ax = plt.subplots(figsize=(7, 3))
mpl_img = ax.imshow(stack[0], cmap="Greys_r")
vals = splines[0](t)
(mpl_line,) = ax.plot(vals[:, 1], vals[:, 0])
ax.set(xlim=(0, stack.shape[2]), ylim=(0, stack.shape[1]))
animation = matplotlib.animation.FuncAnimation(fig, _update2, len(stack), interval=100, blit=True)
plt.show()

# %%
# Analyzing the Undulating Motion
# -------------------------------
# To focus on the undulating motion, we translate the splines to their center of mass and rotate them to align horizontally.

for spline in splines:
    start_point = np.squeeze(spline(0))
    stop_point = np.squeeze(spline(spline.M - 1))
    angle = np.arctan2(*(stop_point - start_point))
    c = np.cos(angle)
    s = np.sin(angle)
    rot_matrix = np.array([[c, -s], [s, c]])
    spline.rotate(rot_matrix)
    com = np.mean(spline(t), axis=0)
    spline.translate(-com)

# %%
# Finally, we visualize the isolated undulating movement and its associated curvature.


def _update3(i):
    vals = splines[i](t)
    curvature = splines[i].curvature(t)
    normals = splines[i].normal(t)
    comb = vals + d * curvature[:, np.newaxis] * normals
    for p, mpl_tooth in zip(ps, mpl_teeth):
        mpl_tooth.set_data([[vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]]])
    mpl_curvature.set_data(comb[:, 1], comb[:, 0])
    mpl_line.set_data(vals[:, 1], vals[:, 0])
    return (*mpl_teeth, mpl_curvature, mpl_line)


fig, ax = plt.subplots(figsize=(7, 3))
vals = splines[0](t)
curvature = splines[0].curvature(t)
normals = splines[0].normal(t)
d = 50
comb = vals + d * curvature[:, np.newaxis] * normals
mpl_teeth = []
ps = np.arange(0, len(comb), 10)
for p in ps:
    mpl_teeth.extend(ax.plot([vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]], color="#568b22"))
(mpl_curvature,) = ax.plot(comb[:, 1], comb[:, 0])
(mpl_line,) = ax.plot(vals[:, 1], vals[:, 0])
ax.set(xlim=(-60, 60), ylim=(-25, 25))
animation = matplotlib.animation.FuncAnimation(fig, _update3, len(stack), interval=100, blit=True)
plt.show()



"""
Spline to mesh
==============
This guide demonstrates how to convert a 3D spline curve into various types of meshes using SplineBox.
"""

# sphinx_gallery_thumbnail_number = 4

import numpy as np
import pyvista
import splinebox

# %%
# 1. Constructing a Spline
# ------------------------
# We begin by creating a circular spline.
M = 4
spline = splinebox.Spline(M=M, basis_function=splinebox.Exponential(M), closed=True)
spline.knots = np.array([[0, 0, 1], [0, 1, 0], [0, 0, -1], [0, -1, 0]])

# %%
# 2. Generating a Mesh Without Radius
# -----------------------------------
# The :code:`step_t` parameter determines the granularity of the resulting mesh, corresponding to the step size in the spline parameter space (t).
# Setting the radius to None or 0 results in a line mesh.

# Generate a simple line mesh
points, connectivity = spline.mesh(step_t=0.1, radius=None)

# Prepend the number of points in each element (2 for a line) for PyVista
connectivity = np.hstack((np.full((connectivity.shape[0], 1), 2), connectivity))

# Create and plot the PyVista mesh
mesh = pyvista.PolyData(points, lines=connectivity)
mesh.plot()

# %%
# 3. Mesh with a Fixed Radius
# ---------------------------
# Here, we generate a surface mesh (a "tube") using a fixed radius.
# We employ the Frenet-Serret frame to avoid selecting an initial vector.

# Generate a surface mesh with a fixed radius
points, connectivity = spline.mesh(step_t=0.1, radius=0.2, frame="frenet")

# Prepend the number of points in each element (3 for triangles) for PyVista
connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

# Create and plot the PyVista mesh
mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# 3. Mesh with an Elliptical Cross-Section
# ----------------------------------------
# You can define a custom cross-section shape by specifying the radius as a function of the spline parameter (:code:`t`) and the polar angle (:code:`phi`).
# Example 1: Elliptical Cross-Section
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


def elliptical_radius(t, phi):
    a = 0.1
    b = 0.05
    phi = np.deg2rad(phi)
    r = (a * b) / np.sqrt((b * np.cos(phi)) ** 2 + (a * np.sin(phi)) ** 2)
    return r


points, connectivity = spline.mesh(step_t=0.1, step_angle=36, radius=elliptical_radius, frame="frenet")

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# Example 2: Varying Radius Along the Spline
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


def radius(t, phi):
    return 0.1 + 0.03 * np.sin(t / spline.M * 16 * np.pi)


points, connectivity = spline.mesh(step_t=0.1, step_angle=36, radius=radius, frame="frenet")

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# 5. Bishop Frame for Mesh Generation
# -----------------------------------
# The Frenet-Serret frame is not defined on straight segments and at inflections point.
# In those cases, we can use the Bishop frame instead. Another advantage of the Bishop frame
# is that it does not twist around the spline.
#
# Correcting Twists with the Bishop Frame
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

# Create a spline with for which the Frenet frame twists
spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=False)
spline.control_points = np.array(
    [
        [0.0, 0.0, 0.0],
        [2.0, 0.0, 0.0],
        [2.0, 2.0, 0.0],
        [2.0, 2.0, 2.0],
        [0.0, 2.0, 2.0],
        [0.0, 0.0, 0.0],
    ]
)

points, connectivity = spline.mesh(step_t=0.1, step_angle=36, radius=elliptical_radius, frame="frenet")

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# You can clearly see how the ellipse twists around the spline.
# The Bishop frame eliminates this twist.

points, connectivity = spline.mesh(step_t=0.1, step_angle=36, radius=elliptical_radius, frame="bishop")

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# Changing the initial vector rotates the frame and therefore the ellipse.

initial_vector = np.array([0.5, -0.5, 1])

points, connectivity = spline.mesh(
    step_t=0.1, step_angle=36, radius=elliptical_radius, frame="bishop", initial_vector=initial_vector
)

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))

mesh = pyvista.PolyData(points, faces=connectivity)
mesh.plot(show_edges=True)

# %%
# 6. Volume Mesh
# --------------
# Finally, you can generate a volumetric mesh by setting the :code:`mesh_type` to :code:`"volume"`.

points, connectivity = spline.mesh(
    radius=radius, step_t=0.5, step_angle=72, initial_vector=initial_vector, mesh_type="volume"
)

connectivity = np.hstack((np.full((connectivity.shape[0], 1), 4), connectivity))
cell_types = np.full(len(connectivity), fill_value=pyvista.CellType.TETRA, dtype=np.uint8)

mesh = pyvista.UnstructuredGrid(connectivity, cell_types, points)
mesh.plot(show_edges=True)

# %%
# For a better understanding of the volume mesh we can explode it.
# This allows us to see the individual tetrahedra.
mesh.explode(factor=0.5).plot(show_edges=True)

# %%
# Tips
# ----
# * Save meshes for visualization in ParaView using :code:`mesh.save("mesh.vtk")`.
# * Surface meshes are open by default. Use the :code:`cap_ends` keyword in :meth:`splinebox.spline_curves.Spline.mesh()` to close them.



"""
Dendrite-Centric Coordinate System
==================================
This example demonstrates how to fit a spline to a dendrite and
align the image coordinate system with the spline.

Data Source: `DeepD3 <https://zenodo.org/records/8428849>`_
Data Used: Crop of DeepD3_Benchmark.tif
"""

# sphinx_gallery_thumbnail_number = 3

import matplotlib.animation
import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import scipy
import skan
import skimage
import splinebox

splinebox_color = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]

# %%
# 1. Load and Inspect the Data
# ----------------------------
# We begin by loading the TIFF data, then visualize the image stack through the z-axis.

img = skimage.io.imread("dendrite.tif")


def _update0(i):
    mpl_img.set_array(img[i])
    return (mpl_img,)


fig, ax = plt.subplots(figsize=(7, 3))
mpl_img = ax.imshow(img[0], cmap="Greys_r", vmin=160, vmax=2000)
ax.set(xlim=(0, img.shape[2]), ylim=(0, img.shape[1]))
animation = matplotlib.animation.FuncAnimation(fig, _update0, len(img), interval=100, blit=True)
plt.show()

# %%
# **Visualize in 3D**
#
# Since we are fitting a 3D spline, let's visualize the image stack in 3D.

grid = pv.ImageData(dimensions=np.array(img.shape) + 1)
grid.origin = (0, 0, 0)
grid.spacing = (1, 1, 1)
grid.cell_data["values"] = img.flatten(order="F")
plotter = pv.Plotter()
plotter.add_volume(grid, cmap="bone", clim=(160, 2000))
plotter.camera_position = "yz"
plotter.show()

# %%
# NOTE: if you don't see the image after switching to 'Interactive Scene' you might have to click and drag on the white space once.
#
# 2. Segmentation and Skeletonization
# -----------------------------------
# We segment the dendrite using Otsu's method and skeletonize it to obtain the pixel coordinates for spline fitting.

thresh = skimage.filters.threshold_otsu(img)
mask = img > thresh
# Keep only the largest connected component
label_img = skimage.measure.label(mask)
label_biggest = np.argmax(np.bincount(label_img.flatten())[1:]) + 1
mask = label_img == label_biggest

# Skeletonize
skeleton = skimage.morphology.skeletonize(mask)

# Get skeleton coordinates
skeleton_points = np.stack(np.where(skeleton), axis=-1)

# Visualize skeleton
skeleton_point_cloud = pv.PolyData(skeleton_points.astype(float))
plotter = pv.Plotter()
plotter.add_volume(grid, cmap="bone", clim=(160, 2000))
plotter.add_mesh(skeleton_point_cloud, color=splinebox_color, point_size=10, render_points_as_spheres=True)
plotter.camera_position = "yz"
plotter.show()

# %%
# **Extract the Longest Path of the Skeleton**
#
# We convert the skeleton into a graph and extract the longest path, which corresponds to the main dendrite.

# Returns a sparse connectivity matrix (graph) and the corresponding pixel coordinates
# for each knot in the graph.
graph, coords = skan.csr.skeleton_to_csgraph(skeleton)
coords = np.stack(coords, axis=-1)


# Use the shortest path algorithm to find the distances between all knots(skeleton points).
dist_matrix, predecessors = scipy.sparse.csgraph.shortest_path(graph, return_predecessors=True)

# Extract the index of the start and end knots of the longest path
start_index, stop_index = np.unravel_index(np.argmax(dist_matrix), dist_matrix.shape)

# Reconstruct the longest path using the predecessor matrix
i = start_index
skeleton_points = []
while i != stop_index:
    skeleton_points.append(coords[i])
    i = predecessors[stop_index, i]
skeleton_points = np.array(skeleton_points)

# Visualize longest path
skeleton_point_cloud = pv.PolyData(skeleton_points.astype(float))
plotter = pv.Plotter()
plotter.add_volume(grid, cmap="bone", clim=(160, 2000))
plotter.add_mesh(skeleton_point_cloud, color=splinebox_color, point_size=10, render_points_as_spheres=True)
plotter.camera_position = "yz"
plotter.show()


# %%
# 3. Fit a Spline
# ---------------
# Now that we have the main points of the dendrite, we fit a spline.

M = 20
basis_function = splinebox.basis_functions.B3()
spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=False)
spline.fit(skeleton_points)

# %%
# 4. Plot the Fitted Spline
# -------------------------
# Let's visualize the spline and its knots along with the segmented dendrite.

# Creat meshes for the spline and the knots of the spline
t = np.linspace(0, M - 1, M * 15)
spline_mesh = pv.MultipleLines(points=spline(t))
knots_point_cloud = pv.PolyData(spline.knots)

# Prepare segmentation mesh
grid = pv.ImageData(dimensions=mask.shape)
mesh = grid.contour([0.5], mask.flatten(order="F"), method="marching_cubes")
mesh = mesh.clean()
mesh = mesh.decimate(0.98)
mesh = mesh.smooth(100)

plotter = pv.Plotter()
plotter.add_mesh(mesh, style="wireframe", color="black")
plotter.add_mesh(spline_mesh, color=splinebox_color, line_width=10)
plotter.add_mesh(knots_point_cloud, color="red", point_size=10, render_points_as_spheres=True)
plotter.camera_position = "yz"
plotter.zoom_camera(2)
plotter.show()

# %%
# 5. Compute Normal Planes
# ------------------------
# To align the image coordinate system with the spline, we compute normal planes along the spline.
# The normal planes are spanned by two vectors, which are normal to the local derivative vector
# of the spline (i.e. the vector pointing in the local direction of the spline).

# Compute derivative vectors along the spline
deriv = spline(t, derivative=1)

# %%
# **Compute the first normal vector**
#
# We select the normal vector that lies in the x-y plane
# (i.e. we set the z component to zero).
normal1 = np.zeros((len(t), 3))
normal1[:, 1] = deriv[:, 2]
normal1[:, 2] = -deriv[:, 1]

# %%
# **Compute the second normal vector**
#
# The second normal vector can be obtained using the cross product.
# This yields a vector that is perpendicular to the two input vectors
# ``deriv`` and ``normal1``.
normal2 = np.zeros((len(t), 3))
normal2 = np.cross(deriv, normal1)

# Normalize vectors
normal1 /= np.linalg.norm(normal1, axis=1)[:, np.newaxis]
normal2 /= np.linalg.norm(normal2, axis=1)[:, np.newaxis]

# %%
# **Visualize Normal Planes**
#
# We scale the vectors for better visibility and plot them to verify they are perpendicular to the spline.
spline_mesh["normal1"] = normal1 * 7
spline_mesh["normal2"] = normal2 * 7

plotter = pv.Plotter()
spline_mesh.set_active_vectors("normal1")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="black")
spline_mesh.set_active_vectors("normal2")
plotter.add_mesh(spline_mesh.arrows, lighting=False, color="red")
plotter.add_mesh(spline_mesh, color=splinebox_color, line_width=10)
plotter.camera_position = "yz"
plotter.zoom_camera(2)
plotter.show()

# %%
# **Extract Pixel Values in Normal Planes**
#
# Finally, we interpolate pixel values from the original image along the computed normal planes.

# Centers of the normal planes
spline_coordinates = spline(t)

# Coefficients for scaling the normal vectors
half_window_size = 25
window_range = np.arange(-half_window_size, half_window_size)
ii, jj = np.meshgrid(window_range, window_range)

# Compute pixel coordinates using scaled normal vectors
normal_planes = np.multiply.outer(ii, normal1) + np.multiply.outer(jj, normal2)

# Fix the order of the axes (spline position first, before the normal directions)
normal_planes = np.rollaxis(normal_planes, 2, 0)

# Position normal planes on spline
normal_planes += spline_coordinates[:, np.newaxis, np.newaxis]

# Interpolate pixel values
shape = normal_planes.shape
vals = scipy.ndimage.map_coordinates(
    img,
    normal_planes.reshape(-1, 3).T,
    order=1,
)
vals = vals.reshape(shape[:-1]).astype(np.float64)

# Mask out pixels outside the volume
mask = (
    (np.min(normal_planes, axis=3) < 0)
    | (normal_planes[:, :, :, 0] > img.shape[0] - 1)
    | (normal_planes[:, :, :, 1] > img.shape[1] - 1)
    | (normal_planes[:, :, :, 2] > img.shape[2] - 1)
)
vals[mask] = np.nan

# %%
# 6. Animate the Dendrite-Centric Image
# -------------------------------------
# We create an animation showing the dendrite as seen along the fitted spline.


def _update1(i):
    mpl_point.set_offsets(
        spline_coordinates[
            i,
            2:0:-1,
        ].T
    )
    mpl_img.set_array(vals[i])
    return (mpl_point, mpl_img)


fig, axes = plt.subplots(1, 2, figsize=(7, 3))
axes[0].imshow(np.max(img, axis=0), cmap="Greys_r", vmin=160, vmax=2000)
mpl_point = axes[0].scatter((spline_coordinates[0, 2],), (spline_coordinates[0, 1],))
mpl_img = axes[1].imshow(vals[0], cmap="Greys_r", vmin=160, vmax=2000)
axes[1].set(xlim=(0, vals.shape[2]), ylim=(0, vals.shape[1]))
axes[1].scatter((half_window_size,), (half_window_size,))
animation = matplotlib.animation.FuncAnimation(fig, _update1, len(vals), interval=100, blit=True)
plt.show()



"""
Plot 3D curvature
-----------------

This example shows how to plot curvature combs in 3D
or how to color the spline based on the local curvature.
Both options are illustrated in pyvista and matplotlib.
"""

import matplotlib.pyplot as plt
import mpl_toolkits
import numpy as np
import pyvista
import splinebox

# %%
# 1. Construct as random spline
# -----------------------------

M = 4
spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=False)
np.random.seed(0)
spline.control_points = np.random.rand(M + 2, 3)

t = np.linspace(0, M - 1, M * 15)

vals = spline(t)

# %%
# 2. Curvature comb with pyvista
# ------------------------------

mesh = pyvista.MultipleLines(points=vals)

curvature = spline.curvature(t)

normals = spline.normal(t, frame="frenet")
normal = -normals[:, 0]

comb = vals + curvature[:, np.newaxis] / 100 * normal

comb_mesh_points = np.zeros((comb.shape[0] * 3 - 1, 3))
comb_mesh_points[0::3] = vals
comb_mesh_points[1::3] = comb
comb_mesh_points[2::3] = comb[1:]
comb_mesh = pyvista.MultipleLines(points=comb_mesh_points)

plotter = pyvista.Plotter()
plotter.add_mesh(mesh, color="blue")
plotter.add_mesh(comb_mesh, color="lightblue")
plotter.show()

# %%
# 3. Curvature comb with matplotlib
# ---------------------------------

ax = plt.figure().add_subplot(projection="3d")
ax.plot(vals[:, 2], vals[:, 1], vals[:, 0], color="blue")
ax.plot(comb_mesh_points[:, 2], comb_mesh_points[:, 1], comb_mesh_points[:, 0], color="lightblue")
plt.show()

# %%
# 4. Colour with pyvista
# ----------------------
#
# We start with a simple line.

points, connectivity = spline.mesh(step_t=t[1] - t[0], radius=None)
connectivity = np.hstack((np.full((connectivity.shape[0], 1), 2), connectivity))
mesh = pyvista.PolyData(points, lines=connectivity)
mesh.cell_data["curvature"] = spline.curvature(t[:-1] + np.diff(t))
mesh.plot()

# %%
# To control the thickness of the the line we can use a mesh

points, connectivity = spline.mesh(step_t=t[1] - t[0], radius=0.01)
connectivity = np.hstack((np.full((connectivity.shape[0], 1), 3), connectivity))
mesh = pyvista.PolyData(points, faces=connectivity)
segment_curvature = spline.curvature(t[:-1] + np.diff(t))
mesh.cell_data["curvature"] = np.repeat(segment_curvature, len(connectivity) / len(segment_curvature))
mesh.plot()

# %%
# 5. Colour with matplotlib
# -------------------------

ax = plt.figure().add_subplot(projection="3d")
segments = np.stack([vals[:-1], vals[1:]], axis=1)
lc = mpl_toolkits.mplot3d.art3d.Line3DCollection(segments)
lc.set_array(segment_curvature)
ax.add_collection(lc)
plt.show()



"""
Distance between splines
------------------------

In this example, we compute the euclidean distance between two splines.
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy
import splinebox

# %%
# For simplicity we create a function that can plot the two splines and the
# distance between the splines at parameter values t_min and s_min.


def plot_splines(spline1, spline2, t_min=None, s_min=None):
    vals1 = spline1(np.linspace(0, spline1.M, 1000))
    vals2 = spline2(np.linspace(0, spline2.M, 1000))
    knots1 = spline1.knots
    knots2 = spline2.knots

    plt.plot(vals1[:, 1], vals1[:, 0])
    plt.plot(vals2[:, 1], vals2[:, 0])
    plt.scatter(knots1[:, 1], knots1[:, 0])
    plt.scatter(knots2[:, 1], knots2[:, 0])

    if t_min is not None and s_min is not None:
        point1 = spline1(t_min)
        point2 = spline2(s_min)
        plt.plot([point1[1], point2[1]], [point1[0], point2[0]], color="k", linestyle="--")

    plt.gca().set_aspect("equal", "box")

    plt.show()


# %%
# We start by constructing two arbitrary closed splines.

basis_function = splinebox.basis_functions.B3()
M = 5
spline1 = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=True)
spline2 = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=True)

spline1.control_points = np.array(
    [
        [1, 2],
        [2, 2],
        [3, 2.5],
        [2.2, 3],
        [1.3, 2.2],
    ]
)

spline2.control_points = np.array(
    [
        [2, -2],
        [2.5, -1],
        [2, -1.5],
        [1.5, -1],
        [1, -2],
    ]
)

# %%
# Plot the splines
plot_splines(spline1, spline2)

# %%
# To get an initial guess of the spline parameter pair with the
# smallest distance, we perform a brute force search.
# Note: This can be made more accurate by increasing the number
# of parameters we interogate.

t = np.linspace(0, spline1.M, 5)
s = np.linspace(0, spline2.M, 5)
vals1 = spline1(t)
vals2 = spline2(s)
distance_vectors = vals1[:, np.newaxis] - vals2[np.newaxis, :]
distances = np.linalg.norm(distance_vectors, axis=-1)
indices = np.unravel_index(np.argmin(distances), distances.shape)
t_min = t[indices[0]]
s_min = s[indices[1]]

plot_splines(spline1, spline2, t_min, s_min)

# %%
# To further refine the estimate we can
# run an optimization.


def distance(parameters):
    val1 = spline1(parameters[0])
    val2 = spline2(parameters[1])
    return np.linalg.norm(val1 - val2)


result = scipy.optimize.minimize(distance, np.array([t_min, s_min]), bounds=((0, spline1.M), (0, spline2.M)))
t_min, s_min = result.x

plot_splines(spline1, spline2, t_min, s_min)



"""
Performance Comparison: splinebox vs scipy
------------------------------------------

In this example, we will compare the performance of ``splinebox`` against ``scipy`` in three tasks:

#. Spline creation
#. Spline evaluation
#. Data fitting

We will measure the time required for each task across different numbers of knots, spline dimensionalities, and whether the spline is open or closed.
"""

import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import splinebox

# Number of times each task is repeated for more reliable results
n_repetitions = 10
# Number of knots
Ms = np.arange(10, 101, 10)

# %%
# Spline Creation and Evaluation
# ------------------------------
# First, we compare the time taken to create splines and evaluate them. We repeat the measurements 11 times, discarding the first due to numba's Just-In-Time (JIT) compilation overhead, which is cached for future runs.

results_creation = []
results_evaluation = []

# Evaluate both open and closed splines
for closed in [True, False]:
    # Dimensionalities: 1D, 2D, 3D, 10D
    for ndim in (1, 2, 3, 10):
        for M in Ms:
            for repetition in range(n_repetitions + 1):
                knots = np.random.rand(M, ndim)

                # Measure splinebox creation time
                start = time.perf_counter_ns()
                spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
                spline.knots = knots
                stop = time.perf_counter_ns()

                if repetition != 0:
                    results_creation.append(
                        [
                            "splinebox",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                # Measure spline evaluation time for different number of parameter values (nt)
                for nt in [10, 100, 1000, 10000]:
                    t = np.linspace(0, M if closed else M - 1, nt)
                    start_eval = time.perf_counter_ns()
                    spline(t)
                    stop_eval = time.perf_counter_ns()

                    if repetition != 0:
                        results_evaluation.append(
                            [
                                "splinebox",
                                f"{ndim}D",
                                M,
                                nt,
                                "closed" if closed else "open",
                                stop - start,
                            ]
                        )

                # Measure scipy spline creation time
                if closed:
                    start = time.perf_counter_ns()
                    t_knots = np.arange(M + 1)
                    knots_periodic = np.concatenate([knots, knots[0][np.newaxis, :]], axis=0)
                    spline = scipy.interpolate.make_interp_spline(t_knots, knots_periodic, k=3, bc_type="periodic")
                    stop = time.perf_counter_ns()
                else:
                    start = time.perf_counter_ns()
                    t_knots = np.arange(M)
                    spline = scipy.interpolate.make_interp_spline(t_knots, knots, k=3, bc_type="natural")
                    stop = time.perf_counter_ns()

                if repetition != 0:
                    results_creation.append(
                        [
                            "scipy",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                # Measure scipy spline evaluation time
                for nt in [10, 100, 1000, 10000]:
                    t = np.linspace(0, M if closed else M - 1, nt)
                    start_eval = time.perf_counter_ns()
                    spline(t)
                    stop_eval = time.perf_counter_ns()
                    if repetition != 0:
                        results_evaluation.append(
                            [
                                "scipy",
                                f"{ndim}D",
                                M,
                                nt,
                                "closed" if closed else "open",
                                stop - start,
                            ]
                        )

# %%
# Plotting Spline Creation Time
# -----------------------------
# Now, that we have collected all of the times, we can turn the results into a data frame and plot the results using seaborn.

# Convert results to DataFrame and plot the creation times
df = pd.DataFrame(results_creation, columns=["Package", "Dimensionality", "Number of knots", "closed", "time [ns]"])
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots()
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Dimensionality",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Creation of {periodicity} splines")
    plt.show()

# %%
# The results show that ``splinbox`` outperforms ``scipy`` in all condition
# for the task of creating a spline from a given set of knots.

# %%
# Plotting Spline Evaluation Time
# -------------------------------

# Convert evaluation results to DataFrame and plot
df = pd.DataFrame(
    results_evaluation,
    columns=["Package", "Dimensionality", "Number of knots", "Number of parameter values", "closed", "time [ns]"],
)
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots(figsize=(7, 7))
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Number of parameter values",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Evaluation of {periodicity} splines")
    plt.show()

# %%
# Once again, ``splinebox`` outperforms ``scipy``, in all conditions for the spline evaluation task.
#
# Data Fitting
# ------------
# Now, we compare the time it takes for ``splinebox`` and ``scipy`` to fit data, i.e., determine control points for a spline with M knots using least-squares fitting.

results_fitting = []

for closed in [True, False]:
    for ndim in (1, 2, 3, 10):
        for M in Ms:
            for repetition in range(n_repetitions + 1):
                data = np.random.rand(M * 10, ndim)

                # Measure splinebox fitting time
                start = time.perf_counter_ns()
                spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
                spline.fit(data)
                stop = time.perf_counter_ns()

                if repetition != 0:
                    results_fitting.append(
                        [
                            "splinebox",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                # Measure scipy fitting time
                k = 3
                if closed:
                    start = time.perf_counter_ns()
                    N = len(data)
                    t = np.arange(-k, M + k + 1)
                    u = np.linspace(0, M, N + 1)[:-1]
                    tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0, per=N)
                    stop = time.perf_counter_ns()

                else:
                    start = time.perf_counter_ns()
                    N = len(data)
                    t = np.arange(-k, M + k)
                    u = np.linspace(0, M - 1, N)
                    tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0)
                    stop = time.perf_counter_ns()

                if repetition != 0:
                    results_fitting.append(
                        [
                            "scipy",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

# %%
# Plotting Fitting Time
# ---------------------
df = pd.DataFrame(
    results_fitting,
    columns=["Package", "Dimensionality", "Number of knots", "closed", "time [ns]"],
)
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots(figsize=(7, 7))
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Dimensionality",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Fitting of {periodicity} splines")
    plt.show()

# %%
# In the fitting task ``splinebox`` is outperformed by ``scipy``'s ``splprep``,
# but is competetive for splines with relatively few knots (i.e. < 20).

# %%
# Results Summary
# ---------------
# #. **Spline Creation:** ``splinebox`` consistently outperforms ``scipy`` across all dimensions for both open and closed splines.
# #. **Spline Evaluation:** Similarly, ``splinebox`` is faster in evaluating splines, especially as the number of parameter values increases.
# #. **Data Fitting:** While ``scipy`` has an edge in fitting tasks, particularly for splines with more than 20 knots, ``splinebox`` remains competitive for smaller splines.
